"""

For debugging test module(s) under lib/

"""

from lib import NemoTests
import IPython

n=NemoTests.NemoTests()

# equD56
#n.setup_equD56()
#n.set_config("configs/equD56_quick.yml")
#n.run_nemo()

# 2008 stuff
#n.setup_south2008()
#n.set_config("configs/PSTest_south2008.yml")
#n.run_nemo()

# quickstart
n.setup_quickstart()
n.set_config("../examples/quickstart/quickstart-clusters.yml")
n.run_nemo()
n.cross_match("testsCache/quickstart-clusters/quickstart-clusters_optimalCatalog.fits", "testsCache/DR5_cluster-catalog_v1.1.fits")
n.check_recovered_ratio("fixed_y_c", "fixed_y_c", toleranceSigma = 1.0, errInKey = "fixed_err_y_c", errOutKey = "fixed_err_y_c",
                        SNRKey = "fixed_SNR",  SNRCut = 5.0,  plotLabel = "quickstart-clusters")
n.status_should_be("SUCCESS")

#IPython.embed()
#sys.exit()

# Point sources
#n.cross_match("inputSourcesCatalog.fits", "configs/PSTest_E-D56/PSTest_E-D56_optimalCatalog.fits")
#n.check_recovered_ratio("I", "deltaT_c", tolerance=0.01, SNRKey="SNR", SNRCut=5.0, plotFileName="plots/amplitudeRecovery.png")
#n.check_recovered_positions(plotFileName="plots/positionRecovery.png")

#c.run_nemo_mass(catalogFileName = "configs/equD56_quick/mocks/mockCatalog_1.fits")



"""

Based on the tile-c script by Mat M.

Here we just want plain-old healpix -> CAR projection, no Galactic coord transform involved,
and no polarization either.

We've added an option to add extra white noise - this is a quick/easy (but suboptimal) way
to avoid ringing in filtered maps (may go away when we move to NSIDE = 8192 sims).

"""

import argparse, os, sys, time
import astropy.io.fits as pyfits
from astropy.wcs import WCS
import IPython
parser = argparse.ArgumentParser()
parser.add_argument("ifiles", nargs="+")
parser.add_argument("template")
parser.add_argument("-L", '--label', type=str, default="CAR") 
parser.add_argument('-W', '--add-white-noise', type=float, default=0)
parser.add_argument("-l", "--lmax",    type=int, default=0)
parser.add_argument("-v", "--verbose", action="store_true")
parser.add_argument("-O", "--outputs", type=str, default="map,ivar")
parser.add_argument('-n', '--npol', type=int, default=1)
args = parser.parse_args()
import numpy as np, healpy
from pixell import enmap, utils, curvedsky, coordinates, mpi, memory

unit  = 1
#euler = np.array([57.06793215,  62.87115487, -167.14056929])*utils.degree
dtype = np.float32
ctype = np.result_type(dtype,0j)
rstep = 100
nside = 0
npol = 1
verbose = args.verbose
outputs = args.outputs.split(",")
if len(outputs) == 0:
    print("No outputs selected - nothing to do")
    sys.exit(0)
t0 = time.time()

def progress(msg):
    if verbose:
        print("%6.2f %6.2f %6.2f %s" % ((time.time()-t0)/60, memory.current()/1024.**3, memory.max()/1024.**3, msg))

comm = mpi.COMM_WORLD

header=pyfits.header.Header.fromtextfile(args.template)
wcs=WCS(header)
shape=(header['NAXIS2'], header['NAXIS1'])
#shape = shape[-2:]
progress("Allocating output map %s %s" % (str((1,)+shape), str(dtype)))
if "map" in outputs:
    for ifile in args.ifiles[comm.rank::comm.size]:
        name = os.path.basename(ifile)
        runit = unit
        #if "545" in name: 
        #	runit *= factor_545
        #	npol = 1
        #elif "857" in name: 
        #	runit *= factor_857
        #	npol = 1
        #else:
        #	npol = 3
        npol=args.npol
        fields = range(npol)
        omap = enmap.zeros((npol,)+shape, wcs, dtype)
        progress("%s TQU read" % name)
        imap  = np.array(healpy.read_map(ifile, fields)).astype(dtype)
        nside = healpy.npix2nside(imap.shape[-1])
        progress("%s TQU mask" % name)
        imap[healpy.mask_bad(imap)] = 0
        progress("%s TQU scale" % name)
        imap *= runit
        nside = healpy.npix2nside(imap.shape[-1])
        lmax  = args.lmax or 3*nside
        progress("%s TQU alm2map" % name)
        alm   = curvedsky.map2alm_healpix(imap, lmax=lmax)
        del imap
        # work around healpix bug
        #progress("%s TQU rotate_alm" % name)
        #alm   = alm.astype(np.complex128,copy=False)
        #healpy.rotate_alm(alm, euler[0], euler[1], euler[2])
        #alm   = alm.astype(ctype,copy=False)
        progress("%s TQU map2alm" % name)
        curvedsky.alm2map_cyl(alm, omap)
        del alm
        ofile = ifile.replace(".fits", "_%s.fits" % (args.label))
        if args.add_white_noise > 0:
            progress("%s adding extra white noise with sigma = %.1f per pixel" % (name, args.add_white_noise))
            omap[0]=omap[0]+np.random.normal(0, args.add_white_noise, omap[0].shape)
        progress("%s TQU write %s" % (name, ofile))
        enmap.write_map(ofile, omap[0])



"""

Just creates the survey mask image from a DS9 .reg file.
    
"""

import os
import sys
import argparse
import astropy.io.fits as pyfits
from astLib import *
import numpy as np
import mahotas.polygon
import IPython

#-------------------------------------------------------------------------------------------------------------
# Main
parser=argparse.ArgumentParser()
parser.add_argument("headerTemplate")
parser.add_argument("regionFile")
args=parser.parse_args()

header=pyfits.header.Header.fromtextfile(args.headerTemplate)
wcs=astWCS.WCS(header, mode = 'pyfits')
shape=(header['NAXIS2'], header['NAXIS1'])

outFileName=args.regionFile.replace(".reg", ".fits")

with open(args.regionFile) as inFile:
    lines=inFile.readlines()
polyList=[]
for line in lines:
    if line.find("polygon") != -1:
        polyPoints=[]
        coords=line.split("polygon(")[-1].split(") ")[0].split(",")
        for i in range(0, len(coords), 2):
            try:
                RADeg, decDeg=[float(coords[i]), float(coords[i+1])]
            except:
                IPython.embed()
                sys.exit()
            x, y=wcs.wcs2pix(RADeg, decDeg)
            polyPoints.append((int(round(y)), int(round(x))))
        polyList.append(polyPoints)

# New method - need for > 2 Gb images
surveyMask=np.zeros(shape, dtype = int)
for polyPoints in polyList:
    mahotas.polygon.fill_polygon(polyPoints, surveyMask)
astImages.saveFITS(outFileName, surveyMask, wcs)




"""

Cross-match cluster catalogs made from individual maps in order to check calibration.

"""

import os
import sys
import glob
import numpy as np
import astropy.table as atpy
from nemo import catalogs
from nemo import plotSettings
from collections import OrderedDict as odict
import pylab as plt
import IPython

#------------------------------------------------------------------------------------------------------------
# Main

# Pick one
#relativeTo='median'
#relativeTo='absolute'
#relativeTo='S16'
relativeTo='S18'

# This only makes sense if relativeTo = 'median', 'S16, or 'S18'
#mode='residualDiff'
mode='ratio'

# Applied to reference catalog only (since now doing forced photometry)
refSNRCut=10.0
print("relativeTo = %s; refSNRCut = %.1f" % (relativeTo, refSNRCut))

plotsDir="plots_%s_%s_%.1f" % (mode, relativeTo, refSNRCut)
os.makedirs(plotsDir, exist_ok = True)

# For testing: multiply ycRef by this
calFactor=1.00
if calFactor != 1.0:
    print("WARNING: calFactor set to %.2f - multiplying ycRef by this factor." % (calFactor))

# Reference catalog - used for cross-match positions across other catalogs
refTab=atpy.Table().read("../MFMF_S18_auto/MFMF_S18_auto_M500.fits")
if relativeTo == 'S16':
    refTab=atpy.Table().read("../MFMF_S16_auto/MFMF_S16_auto_M500.fits")
keepCols=['name', 'RADeg', 'decDeg', 'fixed_SNR', 'fixed_y_c', 'fixed_err_y_c']
removeCols=[]
for k in refTab.keys():
    if k not in keepCols:
        removeCols.append(k)
refTab.remove_columns(removeCols)
refTab=refTab[refTab['fixed_SNR'] > refSNRCut]

# Dictionaries in which to store all cross-matched y_c values
# Each entry will be a list
ycRef={}
ycErrRef={}
yc={}
ycErr={}
labels={}

# Cross match against nemo output
print("Collecting fixed_y_c measurements for each cluster across all maps")
xMatchFiles=glob.glob("outputCatalogs/*.fits")
xMatchFiles.sort()
for f in xMatchFiles:    
    label=os.path.split(f)[-1].split("_optimal")[0]
    print("    %s" % (label))
    tab=atpy.Table().read(f)
    #tab=tab[tab['fixed_SNR'] > SNRCut]
    try:
        refMatched, tabMatched, sep=catalogs.crossMatch(refTab, tab, radiusArcmin = 1.4)
    except:
        raise Exception("Matching probably failed because SNRCut is too low so there were no matches")
    
    for ref, match in zip(refMatched, tabMatched):
        name=ref['name']
        if name not in yc.keys():
            yc[name]=[]
            ycErr[name]=[]
            labels[name]=[]
            ycRef[name]=ref['fixed_y_c']*calFactor
            ycErrRef[name]=ref['fixed_err_y_c']
        yc[name].append(match['fixed_y_c'])
        ycErr[name].append(match['fixed_err_y_c'])
        labels[name].append(label)

nameList=list(yc.keys()); nameList.sort()
for name in nameList:
    yc[name]=np.array(yc[name])
    ycErr[name]=np.array(ycErr[name])

# Check scaled residuals relative to median
# Make a plot for each cluster with a fair number of data points
# Gather a set so we know the typical offset for each map
# We can also make a map of typical offset for each cluster
minPoints=10
resByMap={}
nameList=list(yc.keys()); nameList.sort()
print("Making plots of fixed_y_c for each cluster")
for name in nameList:
    
    # Plot
    if len(yc[name]) < minPoints:
        continue
    if relativeTo == 'absolute':
        res=yc[name]
        resSigma=ycErr[name]
        ylim=None
        ylabel="$\\tilde{y_0}\, (10^{-4})$"
    else:
        if mode == 'relativeDiff':
            if relativeTo == 'median':
                res=(yc[name]-np.median(yc[name]))/ycErr[name]
            else:
                # Relative to S16 or S18 act+planck co-add
                res=(np.array(yc[name])-ycRef[name])/np.sqrt(np.array(ycErr[name])**2+ycErrRef[name]**2)
            resSigma=[1]*len(res)
            ylabel="$\Delta \\tilde{y_0} (\sigma)$"
            ylim=(-4, 4)
        elif mode == 'ratio':
            if relativeTo == 'median':
                res=yc[name]/np.median(yc[name])
                resSigma=(ycErr[name]/yc[name])*res
            else:
                res=np.array(yc[name])/np.array(ycRef[name])
                resSigma=np.sqrt(np.power(ycErr[name]/yc[name], 2)+np.power(ycErrRef[name]/ycRef[name], 2))*res
            ylim=(0, 2)
            ylabel="$\\tilde{y_0} / \\tilde{y_0}$ [%s]" % (relativeTo)
        #medAbsRes=np.median(abs(res))
    
    plotSettings.update_rcParams()
    plt.figure(figsize=(18, 10))
    ax=plt.axes([0.1, 0.4, 0.87, 0.5])
    x=np.arange(1, len(labels[name])+1)
    titleStr=name
    if relativeTo != 'absolute':
        if mode == 'relativeDiff':
            plt.plot(np.linspace(0, len(x)+1, 3), [0]*3, 'k--')
        elif mode == 'ratio':
            plt.plot(np.linspace(0, len(x)+1, 3), [1]*3, 'k--') 
    else:
        # plot median absolute value
        ycMedian=np.median(yc[name])
        plt.plot(np.linspace(0, len(x)+1, 3), [ycMedian]*3, 'k--')
        titleStr=titleStr+" - median %s = %.3f" % (ylabel, ycMedian)
    plt.errorbar(x, res, yerr = [resSigma, resSigma], ecolor = '#AAAAFF', elinewidth = 1.5, fmt = 'D', ms = 6, label = name)
    plt.title(titleStr)#+":   median $\left| \Delta \\tilde{y_0} \\right|$ = %.1f $\sigma$" % (medAbsRes))
    plt.xlim(0, len(labels[name])+1)
    plt.xticks(x, list(labels[name]), rotation = 90)
    plt.ylabel(ylabel)
    if ylim is not None:
        plt.ylim(ylim)
    outPDFName=plotsDir+os.path.sep+name.replace(" ", "_")+".pdf"
    plt.savefig(outPDFName)
    plt.savefig(outPDFName.replace(".pdf", ".png"))
    plt.close()
    print("    %s" % (name))#: median | residual (y0 - median_y_0) | = %.1f sigma" % (name, medAbsRes))
    
    # Gather residuals by map
    for resInMap, mapName in zip(res, labels[name]):
        if mapName not in resByMap:
            resByMap[mapName]=[]
        resByMap[mapName].append(resInMap)

# Plot of mean residual (or ratio) per map +/- std error?
if relativeTo == 'absolute':
    sys.exit()

minPoints=5
medianResByMap=[]
meanResByMap=[]
stdResByMap=[]
stdErrByMap=[]
mapNames=[]
allMapNames=list(resByMap.keys())
allMapNames.sort()
for mapName in allMapNames:
    if len(resByMap[mapName]) < minPoints:
        continue
    medianResByMap.append(np.median(resByMap[mapName]))
    meanResByMap.append(np.mean(resByMap[mapName]))
    stdResByMap.append(np.std(resByMap[mapName]))
    stdErrByMap.append(np.std(resByMap[mapName])/np.sqrt(len(resByMap[mapName])))
    mapNames.append(mapName)
meanResByMap=np.array(meanResByMap)
stdResByMap=np.array(stdResByMap)
stdErrByMap=np.array(stdErrByMap)

#errorBar=stdResByMap
#errorBar=stdErrByMap

plotSettings.update_rcParams()
plt.figure(figsize=(18, 10))
ax=plt.axes([0.1, 0.4, 0.87, 0.5])
x=np.arange(1, len(mapNames)+1)
#plt.title("median $\left| \\rm{mean}\,\Delta \\tilde{y_0} (\sigma) \\right|$ = %.1f" % (np.median(abs(np.array(meanResByMap)))))
titleStr="mean %s by map\nmedian-of-means = %.2f" % (ylabel, np.median(meanResByMap))
if calFactor != 1:
    titleStr=titleStr+" (calFactor = %.3f applied)" % (calFactor)
plt.title(titleStr)
if relativeTo != 'absolute':
    if mode == 'relativeDiff':
        plt.plot(np.linspace(0, len(x)+1, 3), [0]*3, 'k--')
    elif mode == 'ratio':
        plt.plot(np.linspace(0, len(x)+1, 3), [1]*3, 'k--') 
for name, xi in zip(mapNames, x):
    plt.plot([xi]*len(resByMap[name]), resByMap[name], '.', color = '#AAAAAA', zorder = 1)
#ecolor = '#AAAAFF'
plt.errorbar(x, meanResByMap, yerr = [stdResByMap, stdResByMap], ecolor = '#AAAAFF', elinewidth = 3, fmt = 'D', ms = 1,
             zorder = 900)
plt.errorbar(x, meanResByMap, yerr = [stdErrByMap, stdErrByMap], elinewidth = 2, fmt = 'D', ms = 6,
             zorder = 1000)
plt.xlim(0, len(mapNames)+1)
plt.xticks(x, list(mapNames), rotation = 90)
plt.ylabel(ylabel)  # from above...
plt.ylim(ylim)      # from above...
outPDFName=plotsDir+os.path.sep+"meansByMap.pdf"
plt.savefig(outPDFName)
plt.savefig(outPDFName.replace(".pdf", ".png"))
plt.close()

# Text output
outFileName=plotsDir+os.path.sep+"meansByMap.txt"
with open(outFileName, "w") as outFile:
    outFile.write("#mapName\tmean\tsigma\tstdErr\tN\n")
    for mapName, mean, sigma, stdErr in zip(mapNames, meanResByMap, stdResByMap, stdErrByMap):
        outFile.write("%s\t%.3f\t%.3f\t%.3f\t%d\n" % (mapName, mean, sigma, stdErr, len(resByMap[mapName])))
# Wiki table
outFileName=plotsDir+os.path.sep+"wikiMeansByMap.txt"
with open(outFileName, "w") as outFile:
    outFile.write("||border=1 width=80%\n||!mapName ||!mean ||!sigma ||!stdErr ||!N ||\n")
    for mapName, mean, sigma, stdErr in zip(mapNames, meanResByMap, stdResByMap, stdErrByMap):
        outFile.write("|| %s || %.3f || %.3f || %.3f || %d ||\n" % (mapName, mean, sigma, stdErr, len(resByMap[mapName])))

# Individual map histograms on one plot
outPDFName=plotsDir+os.path.sep+"distsByMap.pdf"
binEdges=np.linspace(0, 3, 31) 
numCols=9
numRows=np.ceil(len(mapNames)/numCols)
plotCount=1
plt.figure(figsize=(36, numRows*6))
plt.subplots_adjust(left = 0.02, right = 0.99, bottom = 0.02, top = 0.95, 
                    wspace = 0.2, hspace = 0.2)
for name, mean in zip(mapNames, meanResByMap):
    plt.subplot(numRows, numCols, plotCount)
    plt.hist(resByMap[name], bins = binEdges, density = True)
    yMin, yMax=plt.ylim()
    plt.plot([1]*3, np.linspace(0, yMax, 3), '-')
    plt.plot([mean]*3, np.linspace(0, yMax, 3), ':', label = "%.3f" % (mean))
    plt.ylim(0, yMax)
    #if plotCount > plotCount-numRows:
        #plt.xlabel(ylabel)
    plt.xlim(0, 3)
    plt.legend(loc = 'upper right', fontsize = 8)
    plt.xticks([],[])
    plt.yticks([],[])
    plt.title(name, fontdict = {'size': 9})
    plotCount=plotCount+1
plt.savefig(outPDFName)
plt.savefig(outPDFName.replace(".pdf", ".png"))
plt.close()



"""

This module defines several filter classes, together with a function that uses them to filter maps.

New base classes can be derived from the overall base class :obj:`MapFilter`. There are two main
classes of filter that are currently implemented - :obj:`MatchedFilter` and :obj:`RealSpaceMatchedFilter`.

There are also base classes corresponding to filters with different signal templates (:obj:`BeamFilter`,
:obj:`ArnaudModelFilter`, :obj:`BattagliaModelFilter`). The actual filters that can be used are
derived from these:

* :obj:`BeamMatchedFilter`
* :obj:`ArnaudModelMatchedFilter`
* :obj:`BattagliaModelMatchedFilter`
* :obj:`BeamRealSpaceMatchedFilter`
* :obj:`ArnaudModelRealSpaceMatchedFilter`
* :obj:`BattagliaModelRealSpaceMatchedFilter`

"""

import math
from pixell import enmap
from pixell import fft as enfft
from pixell import powspec
import astropy.wcs as enwcs
from astLib import *
import numpy as np
from numpy import fft
import pylab as plt
import os
from scipy import interpolate
from scipy import ndimage
import astropy.io.fits as pyfits
import astropy.stats as apyStats
import copy
import sys
import glob
import itertools
import nemo
from . import maps
from . import signals
from . import photometry
from . import catalogs
from . import plotSettings
from . import gnfw
from . import completeness
import astropy.table as atpy
import time

#-------------------------------------------------------------------------------------------------------------
def filterMaps(unfilteredMapsDictList, filterParams, tileName, diagnosticsDir = '.', \
               selFnDir = '.', verbose = True, undoPixelWindow = True, useCachedFilter = False, \
               returnFilter = False):
    """Builds and applies filters to the unfiltered map(s). 
    
    Args:
        unfilteredMapsDictList (:obj:`list`): A list of dictionaries, each of which describes a map of the sky
            at some frequency (see :ref:`InputMaps`).
        filterParams (:obj:`dict`): A dictionary containing filter settings (see :ref:`Filters`).
        tileName (:obj:`str`): The name of the tile.
        diagnosticsDir (:obj:`str`, optional): Path to the `diagnostics` directory, where the filters may be
            written to disk.
        selFnDir (:obj:`str`, optional): Path to the `selFn` directory, where area masks may be written.
        verbose (:obj:`bool`, optional): If True, write information about progress to the terminal.
        undoPixelWindow (:obj:`bool`, optional): If True, undo the pixel window effect on the output filtered
            maps.
        useCachedFilter (:obj:`bool`, optional): If True, and a previously made filter is found, it will be
            read from disk, rather than re-calculated (used by source injection simulations).
        returnFilter (:obj:`bool`, optional): If True, the filter object is returned, as well as a dictionary
            containing the filtered map.
    
    Returns:
        A dictionary containing the filtered map in signal units, a signal-to-noise-map, area mask, WCS, and
        a label (for housekeeping).
    
    """
        
    f=filterParams
    label=f['label']+"#"+tileName

    print("... making filtered map %s" % (label))
    filterClass=eval('%s' % (f['class']))
    filterObj=filterClass(f['label'], unfilteredMapsDictList, f['params'], tileName = tileName,
                            diagnosticsDir = diagnosticsDir, selFnDir = selFnDir)
    filteredMapDict=filterObj.buildAndApply(useCachedFilter = useCachedFilter)

    # Keywords we need for photometry later
    filteredMapDict['wcs'].header['BUNIT']=filteredMapDict['mapUnits']
    if 'beamSolidAngle_nsr' in filteredMapDict.keys() and filteredMapDict['beamSolidAngle_nsr'] > 0:
        filteredMapDict['wcs'].header['BEAMNSR']=filteredMapDict['beamSolidAngle_nsr']
        filteredMapDict['wcs'].header['FREQGHZ']=filteredMapDict['obsFreqGHz']
    filteredMapDict['wcs'].updateFromHeader()

    # Undo pixel window function using Sigurd's FFT method (takes into account variable pixel scale etc.)
    # We only need to do this for maps of signal (cancels in S/N map)
    # We do this once because it does take some time...
    # ... and then we can forget about if e.g. stacking or doing forced photometry later
    if undoPixelWindow == True:
        mask=np.equal(filteredMapDict['data'], 0)
        filteredMapDict['data']=enmap.apply_window(filteredMapDict['data'], pow=-1.0)
        filteredMapDict['data'][mask]=0 # just in case we rely elsewhere on zero == no data

    if returnFilter == True:
        return filteredMapDict, filterObj
    
    return filteredMapDict
    
#------------------------------------------------------------------------------------------------------------
class MapFilter(object):
    """Base class for Fourier space filters, defining a common interface.
    
    Args:
        label (:obj:`str`): Unique label for the filter, can be anything.
        unfilteredMapsDictList (:obj:`list`): A list of dictionaries, each of which describes a map of the sky
            at some frequency (see :ref:`InputMaps`).
        paramsDict (:obj:`dict`): Dictionary of filter settings (see :ref:`Filters`).
        tileName (:obj:`str`): Name of the tile in which the filter will be constructed.
        writeFilter (:obj:`bool`): If `True`, save the filter to disk in the `diagnosticsDir` directory.
        forceRebuild (:obj:`bool`): If `True`, rebuild the filter, even a previously calculated filter is
            found cached to disk.
        diagnosticsDir (:obj:`str`, optional): Path to the `diagnostics` directory, where the filters may be
            written to disk.
        selFnDir (:obj:`str`, optional): Path to the `selFn` directory, where area masks may be written.
    
    Attributes:
        label (:obj:`str`): Unique label for the filter, can be anything.
        params (:obj:`dict`): Dictionary of filter settings (see :ref:`Filters`).
        diagnosticsDir (:obj:`str`, optional): Path to the `diagnostics` directory, where the filters may be
            written to disk.
        selFnDir (:obj:`str`, optional): Path to the `selFn` directory, where area masks may be written.       
        tileName (:obj:`str`): Name of the map tile.
        filterFileName (:obj:`str`): Path (under `diagnosticsDir`) where the cached filter is written.
        unfilteredMapsDictList (:obj:`list`): A list of dictionaries, each of which describes a map of the sky
            at some frequency (see :ref:`InputMaps`).
        wcs (:obj:`astWCS.WCS`): Object that contains the map World Coordinate System.
        shape (:obj:`tuple`): Dimensions of the map (height, width) in pixels.
        beamSolidAnglesDict (:obj:`dict`): Dictionary, indexed by map frequency in GHz, holding the beam
            solid angles in nano steradians. Used only for conversion of source amplitudes to flux
            densities in Jy.
    
    """
    def __init__(self, label, unfilteredMapsDictList, paramsDict, tileName = 'PRIMARY', writeFilter = False, 
                 forceRebuild = False, diagnosticsDir = None, selFnDir = None):
        
        self.label=label
        self.params=paramsDict
        
        self.diagnosticsDir=diagnosticsDir
        self.selFnDir=selFnDir
        self.tileName=tileName
        self.filterFileName=self.diagnosticsDir+os.path.sep+self.tileName+os.path.sep+"filter_%s#%s.fits" % (self.label, self.tileName)
        
        # Prepare all the unfilteredMaps (in terms of cutting sections, masks etc.)
        # NOTE: This is a copy to deal with repeated runs (yes, it's necessary)
        self.unfilteredMapsDictList=[]
        for mapDict in unfilteredMapsDictList:
            if 'mapToUse' in self.params.keys() and mapDict['label'] != self.params['mapToUse']:
                continue
            newMapDict=mapDict.copy()
            newMapDict.preprocess(tileName = tileName, diagnosticsDir = diagnosticsDir)
            self.unfilteredMapsDictList.append(newMapDict)
        self.wcs=self.unfilteredMapsDictList[0]['wcs']
        self.shape=self.unfilteredMapsDictList[0]['data'].shape

        # Combine flag masks (yes, we may need to think about this more...)
        self.flagMask=np.zeros(self.shape, dtype = int)
        for mapDict, i in zip(self.unfilteredMapsDictList, range(len(self.unfilteredMapsDictList))):
            self.flagMask=self.flagMask+(mapDict['flagMask']*(i+1))

        # Get beam solid angle info (units: nanosteradians)... we'll need for fluxes in Jy later
        self.beamSolidAnglesDict={}
        for mapDict in self.unfilteredMapsDictList:    
            if 'solidAngle_nsr' in mapDict.keys():
                solidAngle_nsr=mapDict['solidAngle_nsr']
            else:
                beamFileName=mapDict['beamFileName']
                with open(beamFileName, "r") as inFile:
                    lines=inFile.readlines()
                    foundLine=False
                    for line in lines:
                        if line.find("solid angle") != -1:
                            foundLine=True
                            break
                    if foundLine == True:
                        bits=line.split("=")
                        solidAngle_nsr=float(bits[1].split()[0])
                    else:
                        solidAngle_nsr=0.0
            self.beamSolidAnglesDict[mapDict['obsFreqGHz']]=solidAngle_nsr
        
        # For pixell / enmap
        self.enwcs=self.wcs.AWCS

        # We could make this adjustable... added after switch to pixell
        self.apodPix=20
        
        # Check that all maps are the same dimensions
        shape=self.unfilteredMapsDictList[0]['data'].shape
        for mapDict in self.unfilteredMapsDictList:
            if mapDict['data'].shape != shape:
                raise Exception("Maps at different frequencies have different dimensions!")
        
        # This is used by routines that make signal templates
        self.makeRadiansMap()
        
        # This is a default - will get modified (calculated) by buildAndApply
        self.signalNorm=1.0
        self.fRelWeights={}
               
        
    def makeRadiansMap(self):
        """Makes a map of distance in radians from the centre of the map being filtered.
        
        Returns:
            None
        
        """
        
        mapDict=self.unfilteredMapsDictList[0]
        
        # NOTE: int conversion for python3
        x0=int(mapDict['data'].shape[1]/2)
        y0=int(mapDict['data'].shape[0]/2)
        ra0, dec0=self.wcs.pix2wcs(x0, y0)
        ra1, dec1=self.wcs.pix2wcs(x0+1, y0+1)
        self.degPerPixX=astCoords.calcAngSepDeg(ra0, dec0, ra1, dec0)
        self.degPerPixY=astCoords.calcAngSepDeg(ra0, dec0, ra0, dec1)
        
        # Real space map of angular distance from centre in radians, used in making filters and beam
        # NOTE: floor and int conversion added for python3
        xRadRange=np.array([np.arange(int(np.floor(-mapDict['data'].shape[1]/2)), int(mapDict['data'].shape[1]/2), \
                                    dtype=np.float64)*np.radians(self.degPerPixX)]*mapDict['data'].shape[0])
        yRadRange=np.array([np.arange(int(np.floor(-mapDict['data'].shape[0]/2)), int(mapDict['data'].shape[0]/2), \
                                    dtype=np.float64)*np.radians(self.degPerPixY)]*mapDict['data'].shape[1]).transpose()
        rRadRange=np.sqrt(xRadRange**2+yRadRange**2)
        self.radiansMap=rRadRange
        
        
    def buildAndApply(self):
        """Builds and applies the filter to the unfiltered map(s). 
        
        Returns:
            A dictionary, containing the listed keys:
                * data (:obj:`np.ndarray`): The filtered map, in signal units.
                * wcs (:obj:`astWCS.WCS`): WCS object for the map.
                * obsFreqGHz (:obj:`float`): The observing frequency (in GHz) for the map. This is set to `yc` if the output units are set to the central Comptonization parameter.
                * SNMap (:obj:`np.ndarray`): Signal-to-noise map.
                * surveyMask (:obj:`np.ndarray`): Survey mask, where pixels with value 1 indicate valid area that can be searched for objects.
                * mapUnits (:obj:`str`): Either `uK` (for ΔTemperature (μK) with respect to the CMB) or `yc` (for central Comptonization parameter).
                * beamSolidAngle_nsr (:obj:`float`): The beam solid angle in nanosteradians. This is only used for conversion of source amplitudes to flux densities in Jy.
                * label (:obj:`str`): User-defined label for the filter (see :ref:`Filters`).
                * tileName (:obj:`str`): Name of the tile this filtered map corresponds to.
        
        """
        # NOTE: Long lines above to avoid breaking formatting in sphinx
        
        raise Exception("Called a base filter class without a buildAndApply() function implemented.")
        return None
    
    
    def makeForegroundsPower(self):
        """Returns 2d power spectrum in k-space, with the power set by a Planck-like CMB power spectrum.
        
        Returns:
            A 2d :obj:`np.ndarray` containing the noise power.
                       
        """

        # CAMB power spec with Planck 2015 parameters (ish)    
        tab=atpy.Table().read(nemo.__path__[0]+os.path.sep+"data"+os.path.sep+"planck_lensedCls.dat", format = 'ascii')
        tab['TT']=(tab['TT']*2*np.pi)/(tab['L']*(tab['L']+1))
        lmap=enmap.modlmap(self.unfilteredMapsDictList[0]['data'].shape, self.enwcs)
        l2p=interpolate.interp1d(tab['L'], tab['TT'], bounds_error=False, fill_value=0.0)
        fgPower=l2p(lmap)*lmap.shape[0]*lmap.shape[1]
                        
        return fgPower
        

    def makeRealSpaceFilterProfile(self):
        """Makes a 1d real-space profile of the filter, with amplitude normalised to 1 at the frequency of the
        first map given in the unfiltered maps list.
        
        Returns:
            One dimensional profile (:obj:`np.ndarray`), corresponding angular range in arcmin (:obj:`np.ndarray`).
        
        """
        realSpace=fft.ifft2(self.filt).real
        realSpace=fft.fftshift(realSpace)

        # Changed for python3 - we may want to check this...
        # i.e., force to always be int after /2 without this
        x0=int(realSpace.shape[2]/2)
        y0=int(realSpace.shape[1]/2)

        # Arbitrarily normalise to 1 at the maximum (in whichever frequency that occurs)
        normFactor=abs(realSpace[:, y0, x0:]).max() 
        prof=realSpace[:, y0, x0:]/normFactor
        #prof=realSpace[0, y0, x0:]/realSpace[0, y0, x0:].max() # normalise for plot
        arcminRange=np.arange(0, prof.shape[1])*self.degPerPixX*60.0
                
        return prof, arcminRange
    
        
    def saveRealSpaceFilterProfile(self):
        """Saves a real-space profile of the filter as a PNG plot under the `diagnosticsDir` directory.
        
        Returns:
            None
        
        """
        
        prof, arcminRange=self.makeRealSpaceFilterProfile()
        
        # Measure characteristic FWHM
        #tck=interpolate.splrep(arcminRange, prof)
        #FWHMArcmin=interpolate.splev([0.5], tck)*2 # *2 because otherwise would be half width
        
        plotSettings.update_rcParams()
        fig=plt.figure(figsize=(8,8))
        ax=plt.axes([0.14, 0.11, 0.835, 0.86])
        #fig.canvas.set_window_title('Filter Profile in Real Space')
        #plt.title("Filter Profile %s" % (self.label))
        plt.ylabel("Amplitude")
        plt.xlabel("$\\theta$ (arcmin)")
        for row, mapDict in zip(prof, self.unfilteredMapsDictList):
            if mapDict['obsFreqGHz'] is not None:
                label = '%d GHz' % (mapDict['obsFreqGHz'])
            elif mapDict['units'] == 'yc':
                label = 'yc'
            plt.plot(arcminRange, row, label = label)
        plt.xlim(0, 10.0)
        plt.ylim(prof.min(), prof.max()*1.1)
        plt.legend()
        plt.savefig(self.diagnosticsDir+os.path.sep+"realSpaceProfile1d_"+self.label+"#"+self.tileName+".png")
        plt.close()
        
        # Save 2D realspace filter image too
        #astImages.saveFITS(self.diagnosticsDir+os.path.sep+"realSpaceProfile2d_"+self.label+".fits", \
                                #realSpace, wcs)
      

    def makeNoiseMap(self, mapData):
        """Estimate the noise map using local measurements in grid cells, over the whole filtered map
        (see :ref:`Filters` for the config file parameters that control this).
        
        Args:
            mapData (:obj:`np.ndarray`): A filtered map.
        
        Returns:
            Noise map (:obj:`np.ndarray`).
        
        """

        # average all weight maps
        # doesn't matter about spectral weighting, we just want areas with similar characteristics
        # now using this for both 'smart' and original noise grid option
        medWeights=[]
        for mapDict in self.unfilteredMapsDictList:
            medWeights.append(mapDict['weights'])
        medWeights=np.median(np.array(medWeights), axis = 0)
        
        # 'smart option' - measure noise in areas with similar weights (however weights defined)
        if self.params['noiseParams']['noiseGridArcmin'] == "smart":
            try:
                numBins=self.params['noiseParams']['numNoiseBins']
            except:
                raise Exception("Need to give numNoiseBins in noiseParams when using noiseGridArcmin = 'smart'")
            binEdges=np.linspace(medWeights.min(), medWeights.max(), numBins)
            RMSMap=np.zeros(medWeights.shape, dtype = np.float32)
            apodMask=np.not_equal(mapData, 0)
            for i in range(len(binEdges)-1):
                # Find area of similar weight
                binMin=binEdges[i]
                binMax=binEdges[i+1]
                weightMask=np.logical_and(medWeights > binMin, medWeights < binMax)
                # Measure noise in that area from filtered map, and fill in noise map
                chunkValues=mapData[weightMask]
                goodAreaMask=np.greater_equal(apodMask[weightMask], 1.0)
                if 'RMSEstimator' in self.params['noiseParams'].keys() and self.params['noiseParams']['RMSEstimator'] == 'biweight':
                    if goodAreaMask.sum() >= 10:
                        # Astropy version is faster but gives identical results
                        chunkRMS=apyStats.biweight_scale(chunkValues[goodAreaMask], c = 9.0, modify_sample_size = True)
                        #chunkRMS=astStats.biweightScale(chunkValues[goodAreaMask], 6.0)
                    else:
                        chunkRMS=0.
                elif 'RMSEstimator' in self.params['noiseParams'].keys() and self.params['noiseParams']['RMSEstimator'] == 'percentile':
                    chunkRMS=np.percentile(abs(chunkValues[goodAreaMask]), 68.3)
                else:
                    # Default: 3-sigma clipped stdev
                    if np.not_equal(chunkValues, 0).sum() != 0:
                        goodAreaMask=np.greater_equal(apodMask[weightMask], 1.0)
                        chunkMean=np.mean(chunkValues[goodAreaMask])
                        chunkRMS=np.std(chunkValues[goodAreaMask])
                        sigmaClip=3.0
                        for c in range(10):
                            mask=np.less(abs(chunkValues), abs(chunkMean+sigmaClip*chunkRMS))
                            mask=np.logical_and(goodAreaMask, mask)
                            if mask.sum() > 0:
                                chunkMean=np.mean(chunkValues[mask])
                                chunkRMS=np.std(chunkValues[mask])
                    else:
                        chunkRMS=0.
                if chunkRMS > 0:
                    RMSMap[weightMask]=chunkRMS

        # The grid method now recognises numNoiseBins in cells
        else:
            # We may want to just bin and not grid
            if 'noiseGridArcmin' not in self.params['noiseParams'].keys() or self.params['noiseParams']['noiseGridArcmin'] is None:
                overlapPix=0
                numXChunks=1
                numYChunks=1
            else: # The usual gridding
                gridSize=int(round((self.params['noiseParams']['noiseGridArcmin']/60.)/self.wcs.getPixelSizeDeg()))
                overlapPix=int(gridSize/2)
                numXChunks=mapData.shape[1]/gridSize
                numYChunks=mapData.shape[0]/gridSize
            yChunks=np.linspace(0, mapData.shape[0], int(numYChunks+1), dtype = int)
            xChunks=np.linspace(0, mapData.shape[1], int(numXChunks+1), dtype = int)
            apodMask=np.not_equal(mapData, 0)
            RMSMap=np.zeros(mapData.shape, dtype = np.float32)
            # For this mode, interpreted as number of noise bins per cell
            if 'numNoiseBins' in self.params['noiseParams'].keys():
                numBins=self.params['noiseParams']['numNoiseBins']
            else:
                numBins=1
            for i in range(len(yChunks)-1):
                for k in range(len(xChunks)-1):
                    y0=yChunks[i]-overlapPix
                    y1=yChunks[i+1]+overlapPix
                    x0=xChunks[k]-overlapPix
                    x1=xChunks[k+1]+overlapPix
                    if y0 < 0:
                        y0=0
                    if y1 > mapData.shape[0]:
                        y1=mapData.shape[0]
                    if x0 < 0:
                        x0=0
                    if x1 > mapData.shape[1]:
                        x1=mapData.shape[1]
                    chunkValues=mapData[y0:y1, x0:x1]
                    goodAreaMask=np.greater_equal(apodMask[y0:y1, x0:x1], 1.0)
                    if goodAreaMask.sum() == 0:
                        continue
                    # Binning inside cell by weights - to handle sudden noise changes
                    weightValues=medWeights[y0:y1, x0:x1]
                    percentiles=np.arange(0, 100, 100/numBins)
                    binEdges=[]
                    for p in percentiles:
                        binEdges.append(np.percentile(weightValues[goodAreaMask], p))
                    binEdges.append(weightValues[goodAreaMask].max()+1e-6)
                    for b in range(len(binEdges)-1):
                        binMin=binEdges[b]
                        binMax=binEdges[b+1]
                        binMask=np.logical_and(weightValues >= binMin, weightValues < binMax)
                        binValues=chunkValues[binMask*goodAreaMask]
                        if 'RMSEstimator' in self.params['noiseParams'].keys() and self.params['noiseParams']['RMSEstimator'] == 'biweight':
                            if (binMask*goodAreaMask).sum() >= 10:
                                chunkRMS=apyStats.biweight_scale(binValues, c = 9.0, modify_sample_size = True)
                            else:
                                chunkRMS=0.
                        elif 'RMSEstimator' in self.params['noiseParams'].keys() and self.params['noiseParams']['RMSEstimator'] == 'percentile':
                            chunkRMS=np.percentile(abs(binValues), 68.3)
                        else:
                            # Default: 3-sigma clipped stdev
                            if np.not_equal(binValues, 0).sum() != 0:
                                chunkMean=np.mean(binValues)
                                chunkRMS=np.std(binValues)
                                sigmaClip=3.0
                                for c in range(10):
                                    mask=np.less(abs(binValues), abs(chunkMean+sigmaClip*chunkRMS))
                                    if mask.sum() > 0:
                                        chunkMean=np.mean(binValues[mask])
                                        chunkRMS=np.std(binValues[mask])
                            else:
                                chunkRMS=0.
                        if chunkRMS > 0:
                            RMSMap[y0:y1, x0:x1][binMask]=chunkRMS
        
        return RMSMap
    
    
    def loadFRelWeights(self):
        """Reads frequency weights used for relativistic corrections from the filter header.
        
        Returns:
            None
        
        """
        with pyfits.open(self.filterFileName) as img:
            self.fRelWeights={}
            for i in range(1, 10):
                if 'RW%d_GHZ' % (i) in img[0].header.keys():
                    freqGHz=img[0].header['RW%d_GHZ' % (i)]
                    self.fRelWeights[freqGHz]=img[0].header['RW%d' % (i)]


    def makeSignalTemplateMap(self, beam, amplitude = None):
        """Makes a model signal template map. Shape parameters (if applicable) are taken from the object's
        `params` attribute.
        
        Args:
            beam (:class:`nemo.signals.BeamProfile` or str): Either a :class:`nemo.signals.BeamProfile` object,
                or a string that gives the path to a text file that describes the beam profile.
            amplitude (:obj:`float`): Amplitude of the signal template.
        
        Returns:
            Model map (2d :obj:`np.ndarray`).
                
        """
        raise Exception("Called a base filter class without a makeSignalTemplateMap() function implemented.")
        return None
        
        
#------------------------------------------------------------------------------------------------------------
class MatchedFilter(MapFilter):
    """A multi-frequency matched filter, implemented in Fourier space. Derived from :class:`MapFilter`.
    
    """

    def buildAndApply(self, useCachedFilter = False):
        
        fMapsToFilter=[]
        for mapDict in self.unfilteredMapsDictList:
            fMapsToFilter.append(enmap.fft(enmap.apod(mapDict['data'], self.apodPix)))
        fMapsToFilter=np.array(fMapsToFilter)
        
        # NOTE: We've tidied up the config file, so we don't have to feed in surveyMask and psMask like this
        # (see startUp.parseConfig)
        surveyMask=self.unfilteredMapsDictList[0]['surveyMask']
        psMask=self.unfilteredMapsDictList[0]['pointSourceMask']

        if os.path.exists(self.filterFileName) == False and useCachedFilter == False:

            fMapsForNoise=[]
            for i in range(len(self.unfilteredMapsDictList)):
                mapDict=self.unfilteredMapsDictList[i]
                d=mapDict['data']
                if self.params['noiseParams']['method'] == 'dataMap':
                    if 'noiseModelCatalog' in self.params.keys() and self.params['noiseModelCatalog'] is not None:
                        assert(type(self.params['noiseModelCatalog']) == list)
                        for noiseModelCatalog in self.params['noiseModelCatalog']:
                            model=maps.makeModelImage(d.shape, self.wcs, noiseModelCatalog,
                                                      mapDict['beamFileName'],
                                                      obsFreqGHz = mapDict['obsFreqGHz'])
                            if model is not None:
                                d=d-model
                    fMapsForNoise.append(enmap.fft(enmap.apod(d, self.apodPix)))
                elif self.params['noiseParams']['method'] == 'model':
                    # Assuming weights are actually inv var white noise level per pix
                    # (which they are for Sigurd's maps)
                    valid=np.nonzero(mapDict['weights'])
                    RMS=np.mean(1/np.sqrt(mapDict['weights'][valid]))
                    if RMS < 10.0:  # Minimum level to stop this blowing up
                        RMS=10.0
                    # Seeds fixed so that outputs are the same on repeated runs
                    cmb=maps.simCMBMap(self.shape, self.wcs, beam = mapDict['beamFileName'],
                                       seed = 3141592654+i, noiseLevel = RMS)
                    fMapsForNoise.append(enmap.fft(enmap.apod(cmb, self.apodPix)))
                else:
                    raise Exception("'%s' is not a valid filter noise method name - fix the .yml config file" % (self.params['noiseParams']['method']))
            fMapsForNoise=np.array(fMapsForNoise)
        
            # Smoothing noise here is essential
            kernelSize=(3,3)
            noiseCov=[]
            for i in range(len(self.unfilteredMapsDictList)):
                iMap=self.unfilteredMapsDictList[i]
                row=[]
                for j in range(len(self.unfilteredMapsDictList)):
                    jMap=self.unfilteredMapsDictList[j]
                    if self.params['noiseParams']['method'] in ['dataMap', 'model']:
                        NP=np.real(fMapsForNoise[i]*fMapsForNoise[j].conj())
                    elif self.params['noiseParams']['method'] == 'max(dataMap,CMB)':
                        NP=np.real(fMapsForNoise[i]*fMapsForNoise[j].conj())
                        NPCMB=self.makeForegroundsPower() # This needs a beam convolution adding
                        NP=np.maximum.reduce([NP, NPCMB])
                    else:
                        raise Exception("Other noise models not yet re-implemented")
                    NP=ndimage.gaussian_filter(NP, kernelSize)
                    row.append(NP)
                noiseCov.append(row)
            del fMapsForNoise
            noiseCov=np.array(noiseCov)
            
            # Signal frequency weighting
            w=[]
            for mapDict in self.unfilteredMapsDictList:
                if mapDict['units'] != 'yc':
                    # Allows custom weighting in the config file
                    if 'specWeight' in mapDict.keys():
                        w.append(mapDict['specWeight'])
                    # Or standard options (e.g., SZ weighting)
                    else:
                        if self.params['outputUnits'] == 'yc':
                            w.append(signals.fSZ(mapDict['obsFreqGHz']))
                        elif self.params['outputUnits'] == 'uK':
                            # alpha = spectral index (e.g., -0.8 ish for AGN, +3.8 ish for dusty)
                            if 'alpha' in self.params.keys() and self.params['alpha'] is not None:
                                w.append(np.power(mapDict['obsFreqGHz']/self.unfilteredMapsDictList[0]['obsFreqGHz'], 
                                                  self.params['alpha']) )
                            else:
                                w.append(1.0)
                        else:
                            raise Exception('need to specify "outputUnits" ("yc" or "uK") in filter params')
                else:
                    w.append(1.0)   # For TILe-C: there should only be one map if input units are yc anyway...
            w=np.array(w)
                
            # Make FFTs of unit-normalised signal templates for each band
            signalMapsList=[]
            fSignalsArr=[]
            for mapDict in self.unfilteredMapsDictList:
                signalMap=self.makeSignalTemplateMap(mapDict['beamFileName'])
                fSignal=enmap.fft(signalMap)
                signalMapsList.append(signalMap)
                fSignalsArr.append(fSignal)
            fSignalsArr=np.array(fSignalsArr)
                    
            # Build the filter itself
            self.filt=np.zeros([len(self.unfilteredMapsDictList), self.shape[0], self.shape[1]], dtype = np.float32)
            for y in range(0, self.shape[0]):
                for x in range(0, self.shape[1]):
                    try:
                        self.filt[:, y, x]=np.dot(np.linalg.inv(noiseCov[:, :, y, x]), w*abs(fSignalsArr[:, y, x])) 
                    except:
                        continue
            del fSignalsArr
            del noiseCov
                        
            # Use a map with known input signal to figure out how much it has been rolled off by     
            if self.params['outputUnits'] == 'yc':
                # Normalise such that peak value in filtered map == y0, taking out the effect of the beam
                signalMaps=[]
                fSignalMaps=[]
                y0=2e-4
                for mapDict in self.unfilteredMapsDictList:
                    if mapDict['units'] == 'yc':    # For handling TILe-C maps
                        signalMap=self.makeSignalTemplateMap(mapDict['beamFileName'], amplitude = y0)
                    else:                           # The normal case
                        deltaT0=maps.convertToDeltaT(y0, mapDict['obsFreqGHz'])
                        signalMap=self.makeSignalTemplateMap(mapDict['beamFileName'], 
                                                             amplitude = deltaT0)
                    signalMap=enmap.apply_window(signalMap, pow=1.0) # Needed for clusters, 1.5% effect
                    signalMaps.append(signalMap)
                    fSignal=enmap.fft(signalMap)
                    fSignalMaps.append(fSignal)
                signalMaps=np.array(signalMaps)
                fSignalMaps=np.array(fSignalMaps)        
                filteredSignal=self.applyFilter(fSignalMaps)
                # This is a 0.6% difference to the previous version
                cRADeg, cDecDeg=self.wcs.getCentreWCSCoords()
                cx, cy=self.wcs.wcs2pix(cRADeg, cDecDeg)
                mapInterpolator=interpolate.RectBivariateSpline(np.arange(filteredSignal.shape[0]),
                                                                np.arange(filteredSignal.shape[1]),
                                                                filteredSignal, kx = 3, ky = 3)
                peakFilteredSignal=mapInterpolator(cy, cx)[0][0]
                #peakFilteredSignal=filteredSignal.max() # Previous version
                self.signalNorm=y0/peakFilteredSignal
                # For relativistic corrections (see signals module)
                totalSignal=filteredSignal.flatten()[np.argmax(filteredSignal)]
                filteredSignalCube=np.real(enmap.ifft(fSignalMaps*self.filt, normalize = False))
                self.fRelWeights={}
                for filteredSignalPlane, mapDict in zip(filteredSignalCube, self.unfilteredMapsDictList):
                    freqGHz=mapDict['obsFreqGHz']
                    fRelWeight=filteredSignalPlane.flatten()[np.argmax(filteredSignal)]/totalSignal
                    self.fRelWeights[freqGHz]=fRelWeight
                del fSignalMaps
            elif self.params['outputUnits'] == 'uK':
                #if len(self.unfilteredMapsDictList) > 1:
                    #raise Exception("multi-frequency filtering not currently supported for outputUnits 'uK' (point source finding)")
                combinedObsFreqGHz=float(list(self.beamSolidAnglesDict.keys())[0])  # Make less clunky...
                signalMaps=[]
                fSignalMaps=[]
                for mapDict in self.unfilteredMapsDictList:
                    signalMap=self.makeSignalTemplateMap(mapDict['beamFileName'])
                    #signalMap=enmap.apply_window(signalMap, pow=1.0) # Tests with nemoModel confirm not needed here
                    signalMaps.append(signalMap)
                    fSignal=enmap.fft(signalMap)
                    fSignalMaps.append(fSignal)
                signalMaps=np.array(signalMaps)
                fSignalMaps=np.array(fSignalMaps)        
                filteredSignal=self.applyFilter(fSignalMaps)
                self.signalNorm=1.0/filteredSignal.max()
                del fSignalMaps
            else:
                raise Exception('need to specify "outputUnits" ("yc" or "uK") in filter params')
        else:
            print("... loading cached filter")
            self.loadFilter()
            self.params['saveRMSMap']=False
            self.params['saveFilter']=False
            self.params['savePlots']=False
        
        # Apply filter
        filteredMap=self.applyFilter(fMapsToFilter)
        del fMapsToFilter
                
        # Units etc.
        if self.params['outputUnits'] == 'yc':
            mapUnits='yc'
            combinedObsFreqGHz='yc'
            beamSolidAngle_nsr=0.0   # not used for clusters...
        elif self.params['outputUnits'] == 'uK':
            #if len(self.unfilteredMapsDictList) > 1:
                #raise Exception("multi-frequency filtering not currently supported for outputUnits 'uK' (point source finding)")
            combinedObsFreqGHz=float(list(self.beamSolidAnglesDict.keys())[0])  # Make less clunky...
            mapUnits='uK'
            beamSolidAngle_nsr=self.beamSolidAnglesDict[combinedObsFreqGHz]
        else:
            raise Exception('need to specify "outputUnits" ("yc" or "uK") in filter params')
        
        # Apply the point source mask here (before noise estimates etc.)        
        filteredMap=filteredMap*psMask
                
        # Make noise and S/N maps
        RMSMap=self.makeNoiseMap(filteredMap)
        validMask=np.greater(RMSMap, 0)
        SNMap=np.zeros(filteredMap.shape, dtype = np.float32)+filteredMap
        SNMap[validMask]=SNMap[validMask]/RMSMap[validMask]
                
        # Use rank filter to zap edges where RMS will be artificially low - we use a bit of a buffer here
        # NOTE: Now point source mask is applied above, we fill the holes back in here when finding edges
        # NOTE: This all works on maps which have a zero border. If they don't, edgeTrimArcmin has no effect
        if 'edgeTrimArcmin' in self.params.keys() and self.params['edgeTrimArcmin'] > 0:
            trimSizePix=int(round((self.params['edgeTrimArcmin']/60.)/self.wcs.getPixelSizeDeg()))
        elif 'noiseGridArcmin' in self.params['noiseParams'] and self.params['noiseParams']['noiseGridArcmin'] != "smart"\
                and self.params['noiseParams']['noiseGridArcmin'] is not None:
            gridSize=int(round((self.params['noiseParams']['noiseGridArcmin']/60.)/self.wcs.getPixelSizeDeg()))
            trimSizePix=int(round(gridSize*3.0))
        else:
            trimSizePix=0.0
        if trimSizePix  > 0:
            edgeCheck=ndimage.rank_filter(abs(filteredMap+(1-psMask)), 0, size = (trimSizePix, trimSizePix))
            edgeCheck=np.array(np.greater(edgeCheck, 0), dtype = np.float32)
        else:
            edgeCheck=np.ones(filteredMap.shape)
        filteredMap=filteredMap*edgeCheck
        surveyMask=edgeCheck*surveyMask*psMask
        filteredMap=filteredMap*surveyMask # NOTE: Needed for 2-pass (I think)
        del edgeCheck

        # Just in case... we always want to trim the apodized region from the region searched
        # This has no effect if we're using a survey mask already
        # Doing this makes life easier when running tests that use small survey masks or go right to edge of tile otherwise
        apodMask=np.equal(enmap.apod(np.ones(filteredMap.shape), self.apodPix), 1)
        surveyMask=surveyMask*apodMask

        # Apply final survey mask to signal-to-noise map and RMS map
        # NOTE: need to avoid NaNs in here, otherwise map interpolation for e.g. S/N will fail later on
        # NOTE: we now save the mask after detecting objects, as we can detect rings around extremely
        # bright sources and add those to the mask there (see pipelines module)
        SNMap=SNMap*surveyMask
        SNMap[np.isnan(SNMap)]=0.
        RMSMap=RMSMap*surveyMask
                
        if 'savePlots' in self.params and self.params['savePlots'] == True:
            self.saveRealSpaceFilterProfile()
                
        if 'saveFilter' in self.params and self.params['saveFilter'] == True:
            img=pyfits.PrimaryHDU()                                                                                                                                                              
            img.header['SIGNORM']=self.signalNorm
            count=0
            for key in list(self.fRelWeights.keys()):
                count=count+1
                img.header['RW%d_GHZ' % (count)]=key
                img.header['RW%d' % (count)]=self.fRelWeights[key]
            img.data=self.filt                                                                                                                                                                      
            img.writeto(self.filterFileName, overwrite = True) 
            
        # NOTE: What to do about frequency here? Generalise for non-SZ
        return {'data': filteredMap, 'wcs': self.wcs, 'obsFreqGHz': combinedObsFreqGHz, 'SNMap': SNMap,
                'RMSMap': RMSMap, 'surveyMask': surveyMask, 'flagMask': self.flagMask, 'mapUnits': mapUnits,
                'beamSolidAngle_nsr': beamSolidAngle_nsr, 'label': self.label, 'tileName': self.tileName}


    def loadFilter(self):
        """Loads in a previously saved filter.
        
        Returns:
            None
        
        """
        with pyfits.open(self.filterFileName) as img:
            self.filt=img[0].data.astype(np.float32)
            self.signalNorm=img[0].header['SIGNORM']
        self.loadFRelWeights()
        

    def reshapeFilter(self, shape):
        """Uses linear interpolation to transform the filter to the given shape.
        
        Returns:
            Reshaped filter (2d :obj:`np.ndarray`)
        
        """
        
        # If we feed in a 2d shape, make sure we add an axis to keep generalised to multi-frequency
        if len(shape) == 2:
            shape=[self.filt.shape[0], shape[0], shape[1]]
        assert(len(shape) == 3)

        lx, ly=enmap.laxes(self.unfilteredMapsDictList[0]['data'].shape, self.enwcs)
        lxToX=interpolate.interp1d(lx, np.arange(lx.shape[0]), fill_value = 'extrapolate')
        lyToY=interpolate.interp1d(ly, np.arange(ly.shape[0]), fill_value = 'extrapolate')
        lxOut, lyOut=enmap.laxes([shape[1], shape[2]], self.enwcs)
        xOut=lxToX(lxOut)  
        yOut=lyToY(lyOut)
        reshapedFilt=np.zeros(shape, dtype = np.float32)
        for i in range(self.filt.shape[0]):
            filtInterp=interpolate.interp2d(np.arange(ly.shape[0]), np.arange(lx.shape[0]), self.filt[i])
            reshapedFilt[i]=filtInterp(yOut, xOut)
        
        return reshapedFilt
    
            
    def applyFilter(self, mapDataToFilter):
        """Apply the filter to the given map data (must be 3d array, i.e., a cube, with each plane
        corresponding to a different frequency). If the map data is not complex, it will be Fourier
        transformed. If the map data is not the same shape as the filter, the filter will be
        interpolated to match.
        
        Args:
            mapDataToFilter (:obj:`np.ndarray`): A 3d array, where each plane corresponds to a map
                at a different observed frequency. This must match with how the filter was defined
                (see :ref:`Filters` and :ref:`InputMaps`).
            
        Returns:
            Filtered map (2d :obj:`np.ndarray`)
        
        """
        
        # NOTE: need to check appropriate signalNorm after reshaping
        if mapDataToFilter.shape == self.filt.shape:
            filt=self.filt
        else:
            filt=self.reshapeFilter(mapDataToFilter.shape)

        if 'complex' in mapDataToFilter.dtype.name:
            fMapsToFilter=mapDataToFilter
        else:
            fMapsToFilter=enmap.fft(enmap.apod(mapDataToFilter, self.apodPix))

        filteredMap=np.real(enmap.ifft(fMapsToFilter*filt, normalize = False)).sum(axis = 0).astype(np.float32)

        # Optional additional high-pass filter
        if 'bckSub' in self.params.keys() and 'bckSubScaleArcmin' in self.params.keys() and self.params['bckSub'] == True:
            filteredMap=maps.subtractBackground(filteredMap, self.wcs, smoothScaleDeg = self.params['bckSubScaleArcmin']/60.)
        
        filteredMap=filteredMap*self.signalNorm

        return filteredMap
        
#------------------------------------------------------------------------------------------------------------
class RealSpaceMatchedFilter(MapFilter):
    """Makes a matched-filter kernel using the noise properties of a specified region of the map (e.g.,
    the deepest part) in Fourier space, which is converted into real space and truncated such that the 
    kernel is small enough in footprint to be applied by direct convolution with the map in a (relatively)
    short amount of time. Derived from :class:`MapFilter`.
        
    """       
    
    def loadFilter(self):
        """Loads in a previously saved filter kernel.
        
        Returns:
            None
        
        """
        
        with pyfits.open(self.filterFileName) as img:
            kern2d=img[0].data
            signalNorm=img[0].header['SIGNORM']
            if 'BCKSCALE' in img[0].header.keys():
                bckSubScaleArcmin=img[0].header['BCKSCALE']
            else:
                bckSubScaleArcmin=0
            self.applyRACentre=img[0].header['APP_RA']
            self.applyDecCentre=img[0].header['APP_DEC']
        self.kern2d=kern2d
        self.signalNorm=signalNorm
        self.bckSubScaleArcmin=bckSubScaleArcmin

        
    def buildKernel(self, RADecSection, RADeg = 'centre', decDeg = 'centre'):
        """Builds the real space filter kernel.
        
        Args:
            RADecSection (:obj:`list`): List of coordinates (minimum RA, maximum RA,
                minimum declination, maximum declination) that defines the region of map
                from which the filter kernel will be constructed.
            RADeg (:obj:`str` or :obj:`float`, optional): Coordinate at which the kernel
                pixel scale will be determined (use `centre` for the middle of the map
                section).
            decDeg (:obj:`str` or :obj:`float`, optional): Coordinate at which the kernel
                pixel scale will be determined (use `centre` for the middle of the map
                section).
                
        Returns:
            None
            
        """
        
        if os.path.exists(self.filterFileName) == True:
            return self.loadFilter()
        
        wcs=self.wcs
        
        # Build the matched-filter kernel in a small section of the map
        # Apply the same difference of Gaussians high pass filter here
        # NOTE: we could merge 'bckSubScaleArcmin' and 'maxArcmin' keys here!
        #mapDict['bckSubScaleArcmin']=maxArcmin
        keysWanted=['mapFileName', 'weightsFileName', 'obsFreqGHz', 'units', 'beamFileName', 'addNoise', 
                    'pointSourceRemoval', 'weightsType', 'tileName']
        kernelUnfilteredMapsDictList=[]
        for mapDict in self.unfilteredMapsDictList:
            for key in list(mapDict.keys()):
                if key not in keysWanted:
                    del mapDict[key]
            kernelUnfilteredMapsDict=maps.MapDict(mapDict, tileCoordsDict = mapDict.tileCoordsDict)
            kernelUnfilteredMapsDict['RADecSection']=RADecSection
            kernelUnfilteredMapsDictList.append(kernelUnfilteredMapsDict)
        kernelLabel="realSpaceKernel_%s" % (self.label)
        matchedFilterDir=self.diagnosticsDir+os.path.sep+kernelLabel+"#"+self.tileName
        diagnosticsDir=matchedFilterDir+os.path.sep+'diagnostics'
        selFnDir=matchedFilterDir+os.path.sep+'selFn'
        for d in [matchedFilterDir, diagnosticsDir, selFnDir]:
            if os.path.exists(d) == False:
                os.makedirs(d, exist_ok = True)
        matchedFilterClass=eval(self.params['noiseParams']['matchedFilterClass'])
        matchedFilter=matchedFilterClass(kernelLabel, kernelUnfilteredMapsDictList, self.params,
                                         tileName = mapDict['tileName'],
                                         diagnosticsDir = matchedFilterDir+os.path.sep+'diagnostics',
                                         selFnDir = matchedFilterDir+os.path.sep+'selFn')
        filteredMapDict=matchedFilter.buildAndApply()
        
        # Turn the matched filter into a smaller real space convolution kernel
        # This means we have to roll off the kernel to 0 at some radius
        # This is set by maxArcmin in the config file
        kernelMaxArcmin=self.params['noiseParams']['kernelMaxArcmin']
        prof, arcminRange=matchedFilter.makeRealSpaceFilterProfile()
        rIndex=np.where(arcminRange > kernelMaxArcmin)[0][0]
        mask=np.less(arcminRange, kernelMaxArcmin)

        # Kernel can be either fully 2d, or be azimuthally averaged... in the ACTPol E-D56 paper, we used the latter
        if 'symmetrize' in self.params['noiseParams'].keys() and self.params['noiseParams']['symmetrize'] == True:
            rRadians=np.radians(arcminRange/60.)
            profile2d=[]
            for i in range(prof.shape[0]):
                r2p=interpolate.interp1d(rRadians[mask], prof[i, mask], bounds_error=False, fill_value=0.0)
                profile2d.append(r2p(matchedFilter.radiansMap))
            profile2d=np.array(profile2d)
        else:
            profile2d=fft.ifft2(matchedFilter.filt).real
            profile2d=fft.fftshift(profile2d)
            
        # z is not needed here - just because we switched to multi-freq throughout
        z, y, x=np.where(abs(profile2d) == abs(profile2d).max()) 
        #y, x=np.where(profile2d[0] == profile2d[0].max())
        y=y[0]
        x=x[0]
        yMin=y-rIndex
        yMax=y+rIndex
        xMin=x-rIndex
        xMax=x+rIndex
        if (yMax-yMin) % 2 == 0:
            yMin=yMin+1
        if (xMax-xMin) % 2 == 0:
            xMin=xMin+1
        self.kern2d=profile2d[:, yMin:yMax, xMin:xMax]
        kern2dRadiansMap=matchedFilter.radiansMap[yMin:yMax, xMin:xMax]

        # This is what to high pass filter on
        if 'bckSubScaleArcmin' in self.params.keys():
            self.bckSubScaleArcmin=self.params['bckSubScaleArcmin']
        else:
            # This is here just so we can reproduce old results where this was done automatically
            # Now we have to fiddle about a bit to check the sign of the filter 
            # (depends on spectral response)
            # We're only considering the first frequency given in a set of multi-freq maps
            if np.greater(prof[0, 0], 0) == True:
                func=np.min
            else:
                func=np.max
            self.bckSubScaleArcmin=arcminRange[prof[0] == func(prof[0])][0]
            
        # Use a map with known input signal to figure out how much it has been rolled off by:
        # 1. The high pass filter (bck sub step)
        # 2. The matched filter itself (includes beam)       
        # NOTE: This is SZ-specific again and needs generalising
        signalMaps=[]
        for mapDict in self.unfilteredMapsDictList:
            # This has been made more complicated because of TILe-C
            if self.params['outputUnits'] == 'yc':
                y0=2e-4
                if mapDict['obsFreqGHz'] is not None:   
                    # Normal case
                    deltaT0=maps.convertToDeltaT(y0, mapDict['obsFreqGHz'])
                    signalMap=self.makeSignalTemplateMap(mapDict['beamFileName'],
                                                         amplitude = deltaT0)
                else:
                    # TILe-C case
                    signalMap=self.makeSignalTemplateMap(mapDict['beamFileName'], amplitude = y0)
            elif self.params['outputUnits'] == 'uK':
                signalMap=self.makeSignalTemplateMap(mapDict['beamFileName'])
            else:
                raise Exception('need to specify "outputUnits" ("yc" or "uK") in filter params')
            signalMaps.append(signalMap)
        signalMaps=np.array(signalMaps)
        
        filteredSignal=self.applyFilter(signalMaps, calcFRelWeights = True)
        if self.params['outputUnits'] == 'yc':
            # Normalise such that peak value in filtered map == y0, taking out the effect of the beam
            self.signalNorm=y0/filteredSignal.max()            
        elif self.params['outputUnits'] == 'uK':
            self.signalNorm=1.0/filteredSignal.max()
        else:
            raise Exception('need to specify "outputUnits" ("yc" or "uK") in filter params')
        
        # Save 2d kernel - we need this (at least for the photometry ref scale) to calc Q later
        # Add bckSubScaleArcmin to the header
        kernWCS=wcs.copy()
        if self.params['bckSub'] == True:
            kernWCS.header['BCKSCALE']=self.bckSubScaleArcmin
        kernWCS.header['SIGNORM']=self.signalNorm
        kernWCS.header['APP_RA']=self.applyRACentre
        kernWCS.header['APP_DEC']=self.applyDecCentre
        count=0
        for key in list(self.fRelWeights.keys()):
            count=count+1
            kernWCS.header['RW%d_GHZ' % (count)]=key
            kernWCS.header['RW%d' % (count)]=self.fRelWeights[key]
        kernWCS.header['NEMOVER']=nemo.__version__
        maps.saveFITS(self.filterFileName, self.kern2d, kernWCS)
        
        # Filter profile plot   
        # Save the stuff we plot first, in case we want to make a plot with multiple filters on later
        np.savez(self.diagnosticsDir+os.path.sep+"filterProf1D_%s#%s.npz" % (self.label, self.tileName), 
                 arcminRange = arcminRange, prof = prof, mask = mask, 
                 bckSubScaleArcmin = self.bckSubScaleArcmin)
        plotSettings.update_rcParams()
        plt.figure(figsize=(9,6.5))
        ax=plt.axes([0.13, 0.12, 0.86, 0.86])
        #plt.tick_params(axis='both', which='major', labelsize=15)
        #plt.tick_params(axis='both', which='minor', labelsize=15)
        for row, mapDict in zip(prof, self.unfilteredMapsDictList):
            tck=interpolate.splrep(arcminRange[mask], row[mask])
            plotRange=np.linspace(0, arcminRange[mask].max(), 1000)
            #plt.plot(arcminRange[mask], prof[mask])
            if mapDict['obsFreqGHz'] is not None:
                label = '%d GHz' % (mapDict['obsFreqGHz'])
            elif mapDict['units'] == 'yc':
                label = 'yc'
            plt.plot(plotRange, interpolate.splev(plotRange, tck), '-', label = label)
        plt.xlabel("$\\theta$ (arcmin)")
        plt.ylabel("Amplitude")
        plt.legend()
        #plt.title(self.label)
        #plt.plot(arcminRange[mask], [0]*len(arcminRange[mask]), 'k--')
        plt.xlim(0, arcminRange[mask].max())
        if self.params['bckSub'] == True:
            plt.plot([self.bckSubScaleArcmin]*3, np.linspace(-1.2, 1.2, 3), 'k--')
        plt.ylim(-1.2, 0.2)
        plt.savefig(self.diagnosticsDir+os.path.sep+"filterPlot1D_%s#%s.pdf" % (self.label, self.tileName))
        plt.close()

            
    def buildAndApply(self):

        surveyMask=self.unfilteredMapsDictList[0]['surveyMask']
        psMask=self.unfilteredMapsDictList[0]['pointSourceMask']

        if os.path.exists(self.filterFileName) == False:

            # Noise region to use
            RAMin, RAMax, decMin, decMax=self.wcs.getImageMinMaxWCSCoords()
            if self.params['noiseParams']['RADecSection'] == 'tileNoiseRegions':
                RADecSection=[self.wcs.header['NRAMIN'], self.wcs.header['NRAMAX'], 
                              self.wcs.header['NDEMIN'], self.wcs.header['NDEMAX']]
            elif self.params['noiseParams']['RADecSection'] == 'auto':
                cRA, cDec=self.wcs.getCentreWCSCoords()
                halfSizeDeg=2.0
                nRAMin=cRA-halfSizeDeg/np.cos(np.radians(cDec))
                nRAMax=cRA+halfSizeDeg/np.cos(np.radians(cDec))
                nDecMin=cDec-halfSizeDeg
                nDecMax=cDec+halfSizeDeg
                RADecSection=[nRAMin, nRAMax, nDecMin, nDecMax]
            else:
                RADecSection=self.params['noiseParams']['RADecSection']
            self.applyDecCentre=(decMax+decMin)/2.
            self.applyRACentre=(RAMax+RAMin)/2.
            
            # Build kernel   
            self.buildKernel(RADecSection, RADeg = self.applyRACentre, decDeg = self.applyDecCentre)
        else:
            self.loadFilter()

        # Apply kernel
        mapDataToFilter=[]
        for mapDict in self.unfilteredMapsDictList:
            mapDataToFilter.append(mapDict.loadTile('mapFileName', tileName = self.tileName))
        mapDataToFilter=np.array(mapDataToFilter)
        filteredMap=self.applyFilter(mapDataToFilter)
                                                
        # Apply the point source mask here (before noise estimates etc.)
        filteredMap=filteredMap*psMask
        
        # Make noise and S/N maps
        RMSMap=self.makeNoiseMap(filteredMap)
        validMask=np.greater(RMSMap, 0)
        SNMap=np.zeros(filteredMap.shape, dtype = np.float32)+filteredMap
        SNMap[validMask]=SNMap[validMask]/RMSMap[validMask]
        
        # Units etc.
        if self.params['outputUnits'] == 'yc':
            mapUnits='yc'
            combinedObsFreqGHz='yc'
            beamSolidAngle_nsr=0.0   # not used for clusters...
        elif self.params['outputUnits'] == 'uK':
            if len(self.unfilteredMapsDictList) > 1:
                raise Exception("multi-frequency filtering not currently supported for outputUnits 'uK' (point source finding)")
            combinedObsFreqGHz=float(list(self.beamSolidAnglesDict.keys())[0])  # Make less clunky...
            mapUnits='uK'
            beamSolidAngle_nsr=self.beamSolidAnglesDict[combinedObsFreqGHz]
        else:
            raise Exception('need to specify "outputUnits" ("yc" or "uK") in filter params')

        # Use rank filter to zap edges where RMS will be artificially low - we use a bit of a buffer here
        # NOTE: Now point source mask is applied above, we fill the holes back in here when finding edges
        if 'edgeTrimArcmin' in self.params.keys():
            trimSizePix=int(round((self.params['edgeTrimArcmin']/60.)/self.wcs.getPixelSizeDeg()))
        else:
            gridSize=int(round((self.params['noiseParams']['noiseGridArcmin']/60.)/self.wcs.getPixelSizeDeg()))
            trimSizePix=int(round(gridSize*3.0))
        if trimSizePix > 0:
            edgeCheck=ndimage.rank_filter(abs(filteredMap+(1-psMask)), 0, size = (trimSizePix, trimSizePix))
        else:
            edgeCheck=np.ones(filteredMap.shape)
        edgeCheck=np.array(np.greater(edgeCheck, 0), dtype = float)
        filteredMap=filteredMap*edgeCheck
        surveyMask=edgeCheck*surveyMask*psMask
        del edgeCheck

        # Just in case... we always want to trim the apodized region from the region searched
        # This has no effect if we're using a survey mask already
        # Doing this makes life easier when running tests that use small survey masks or go right to edge of tile otherwise
        apodMask=np.equal(enmap.apod(np.ones(filteredMap.shape), self.apodPix), 1)
        surveyMask=surveyMask*apodMask

        # Apply final survey mask to signal-to-noise map and RMS map
        # NOTE: need to avoid NaNs in here, otherwise map interpolation for e.g. S/N will fail later on
        SNMap=SNMap*surveyMask
        SNMap[np.isnan(SNMap)]=0.
        RMSMap=RMSMap*surveyMask

        return {'data': filteredMap, 'wcs': self.wcs, 'obsFreqGHz': combinedObsFreqGHz, 'SNMap': SNMap, 
                'RMSMap': RMSMap, 'surveyMask': surveyMask, 'flagMask': self.flagMask, 'mapUnits': mapUnits,
                'beamSolidAngle_nsr': beamSolidAngle_nsr, 'label': self.label, 'tileName': self.tileName}

    
    def applyFilter(self, mapDataToFilter, calcFRelWeights = False):
        """Apply the filter to the given map data (must be 3d array, i.e., a cube, with each plane
        corresponding to a different frequency).

        Args:
            mapDataToFilter (:obj:`np.ndarray`): A 3d array, where each plane corresponds to a map
                at a different observed frequency. This must match with how the filter was defined
                (see :ref:`Filters` and :ref:`InputMaps`).
            calcFRelWeights (:obj:`bool`, optional): This should *only* be set to `True` if this
                routine is being applied to an ideal signal map.
            
        Returns:
            Filtered map (2d :obj:`np.ndarray`)
        
        """
        
        # Apply the high pass filter - subtract background on larger scales using difference of Gaussians 
        filteredMap=np.zeros(mapDataToFilter.shape, dtype = np.float32)
        if self.params['bckSub'] == True and self.bckSubScaleArcmin > 0:
            for i in range(mapDataToFilter.shape[0]):
                filteredMap[i]=maps.subtractBackground(mapDataToFilter[i], self.wcs, 
                                                       RADeg = self.applyRACentre,
                                                       decDeg = self.applyDecCentre,
                                                       smoothScaleDeg = self.bckSubScaleArcmin/60.)
        else:
            filteredMap=filteredMap+mapDataToFilter
        
        # Apply the kernel
        for i in range(filteredMap.shape[0]):
            filteredMap[i]=ndimage.convolve(filteredMap[i], self.kern2d[i])
        
        # For relativistic corrections (see signals module)
        if calcFRelWeights == True:
            self.fRelWeights={}
            maxIndex=np.argmax(filteredMap.sum(axis = 0))
            totalSignal=filteredMap.sum(axis = 0).flatten()[maxIndex]
            for filteredSignalPlane, mapDict in zip(filteredMap, self.unfilteredMapsDictList):
                freqGHz=mapDict['obsFreqGHz']
                fRelWeight=filteredSignalPlane.flatten()[maxIndex]/totalSignal
                self.fRelWeights[freqGHz]=fRelWeight
                
        filteredMap=filteredMap.sum(axis = 0).astype(np.float32)
        
        # Apply the normalisation
        filteredMap=filteredMap*self.signalNorm
        
        return filteredMap
    
        
#------------------------------------------------------------------------------------------------------------
class BeamFilter(MapFilter):
    """Base class for filters using ACT-style beam profile files. Derived from :class:`MapFilter`.
        
    """
    
    def makeSignalTemplateMap(self, beamFileName, amplitude = None):
        signalMap=signals.makeBeamModelSignalMap(np.degrees(self.radiansMap),
                                                            self.wcs, 
                                                            beamFileName,
                                                            amplitude = amplitude)

        return signalMap
    
#------------------------------------------------------------------------------------------------------------
class ArnaudModelFilter(MapFilter):
    """Base class for filters using the GNFW profile as described in 
    `Arnaud et al. (2010) <https://ui.adsabs.harvard.edu/abs/2010A%26A...517A..92A/abstract>`_.
    Derived from :class:`MapFilter`.
    
    """
    
    def makeSignalTemplateMap(self, beamFileName, amplitude = None):
        RADeg, decDeg=self.wcs.getCentreWCSCoords()
        signalMap=signals.makeArnaudModelSignalMap(self.params['z'], self.params['M500MSun'],
                                                   self.shape, self.wcs, beam = beamFileName,
                                                   RADeg = RADeg, decDeg = decDeg,
                                                   GNFWParams = self.params['GNFWParams'],
                                                   amplitude = amplitude,
                                                   convolveWithBeam = True)

        return signalMap

#------------------------------------------------------------------------------------------------------------
class BattagliaModelFilter(MapFilter):
    """Base class for filters using the GNFW profile as described in
    `Battaglia et al. (2012) <https://ui.adsabs.harvard.edu/abs/2012ApJ...758...75B/abstract>`_.
    Derived from :class:`MapFilter`.
    
    Note:
        This is the same as :class:`ArnaudModelFilter` (it is still a GNFW profile), but has non-self-similar
        evolution with redshift, which is accounted for here. The convention for how the GNFW parameters are
        defined follows `Arnaud et al. (2010) <https://ui.adsabs.harvard.edu/abs/2010A%26A...517A..92A/abstract>`_,
        rather than `Battaglia et al. (2012) <https://ui.adsabs.harvard.edu/abs/2012ApJ...758...75B/abstract>`_.

    """
    
    def makeSignalTemplateMap(self, beamFileName, amplitude = None):
        RADeg, decDeg=self.wcs.getCentreWCSCoords()
        signalMap=signals.makeBattagliaModelSignalMap(self.params['z'], self.params['M500MSun'],
                                                   self.shape, self.wcs, beam = beamFileName,
                                                   RADeg = RADeg, decDeg = decDeg,
                                                   GNFWParams = self.params['GNFWParams'],
                                                   amplitude = amplitude,
                                                   convolveWithBeam = True)
        
        return signalMap
    
#------------------------------------------------------------------------------------------------------------
# Definitions of actual filters that can be used

class ArnaudModelMatchedFilter(MatchedFilter, ArnaudModelFilter):
    """Fourier-space multi-frequency matched filter implementation, using the GNFW profile as described in 
    `Arnaud et al. (2010) <https://ui.adsabs.harvard.edu/abs/2010A%26A...517A..92A/abstract>`_.
    Derived from :class:`MapFilter`.
    
    """
    pass

class BattagliaModelMatchedFilter(MatchedFilter, BattagliaModelFilter):
    """Fourier-space multi-frequency matched filter implementation, using the GNFW profile as described in
    `Battaglia et al. (2012) <https://ui.adsabs.harvard.edu/abs/2012ApJ...758...75B/abstract>`_.
    Derived from :class:`MapFilter`.
    
    Note:
        This is the same as :class:`ArnaudModelFilter` (it is still a GNFW profile), but has non-self-similar
        evolution with redshift, which is accounted for here. The convention for how the GNFW parameters are
        defined follows `Arnaud et al. (2010) <https://ui.adsabs.harvard.edu/abs/2010A%26A...517A..92A/abstract>`_,
        rather than `Battaglia et al. (2012) <https://ui.adsabs.harvard.edu/abs/2012ApJ...758...75B/abstract>`_.
        
    """
    pass

class BeamMatchedFilter(MatchedFilter, BeamFilter):
    """Fourier-space multi-frequency matched filter implementation using ACT-style beam profile files. 
    Derived from :class:`MapFilter`.
        
    """
    pass

class ArnaudModelRealSpaceMatchedFilter(RealSpaceMatchedFilter, ArnaudModelFilter):
    """Real-space filter kernel, built using the GNFW profile as described in
    `Arnaud et al. (2010) <https://ui.adsabs.harvard.edu/abs/2010A%26A...517A..92A/abstract>`_.
    Derived from :class:`MapFilter`.
    
    """
    pass

class BattagliaModelRealSpaceMatchedFilter(RealSpaceMatchedFilter, BattagliaModelFilter):
    """Real-space filter kernel, built using the GNFW profile as described in
    `Battaglia et al. (2012) <https://ui.adsabs.harvard.edu/abs/2012ApJ...758...75B/abstract>`_.
    Derived from :class:`MapFilter`.
    """
    pass

class BeamRealSpaceMatchedFilter(RealSpaceMatchedFilter, BeamFilter):
    """Real-space filter kernel, built using ACT-style beam profile files. 
    Derived from :class:`MapFilter`.
    
    """
    pass



"""

This module defines the MockSurvey class, used for mass function calculations, obtaining de-biased cluster
mass estimates, selection function calculations, and generating mock catalogs.

"""

import os
import sys
import numpy as np
import astropy.table as atpy
import pylab as plt
import subprocess
from astropy.cosmology import FlatLambdaCDM
on_rtd=os.environ.get('READTHEDOCS', None)
if on_rtd is None:
    import pyccl as ccl
from . import signals
from . import catalogs
from . import maps
import pickle
from scipy import interpolate
from scipy import integrate
from scipy.interpolate import InterpolatedUnivariateSpline as _spline
from scipy import stats
from astLib import *
import time

from classy_sz import Class

common_class_sz_settings = {
                   'mass function' : 'T08M200c', 
                   'hm_consistency': 0,
                   'concentration parameter' : 'B13',
                   'B':1.,

                   'N_ncdm' : 1,
                   'N_ur' : 2.0328,
                   'm_ncdm' : 0.0,
                   'T_ncdm' : 0.71611,
    
                   'z_min': 1.e-3, # chose a wide range 
                   'z_max': 3.05, # chose a wide range 
                   'redshift_epsrel': 1e-6,
                   'redshift_epsabs': 1e-100,  
    
                   'M_min': 1e13, # chose a wide range 
                   'M_max': 1e17, # chose a wide range 
                   'mass_epsrel':1e-6,
                   'mass_epsabs':1e-100,
   
                   'ndim_redshifts' :5000,
                   'ndim_masses' : 500,
                   'n_m_dndlnM' : 1000,
                   'n_z_dndlnM' : 1000,
                   'HMF_prescription_NCDM': 1,
                   'no_spline_in_tinker': 1,
    
    
                   'use_m500c_in_ym_relation' : 0,
                   'use_m200c_in_ym_relation' : 1,
                   'y_m_relation' : 1,
    
                   'output': 'dndlnM,m500c_to_m200c,m200c_to_m500c',
}


#------------------------------------------------------------------------------------------------------------
class MockSurvey(object):
    """An object that provides routines calculating cluster counts (using `CCL <https://ccl.readthedocs.io/en/latest/>`_) 
    and generating mock catalogs for a given set of cosmological and mass scaling relation parameters.
    The Tinker et al. (2008) halo mass function is used (hardcoded at present, but in principle this can
    easily be swapped for any halo mass function supported by CCL).
    
    Attributes:
        areaDeg2 (:obj:`float`): Survey area in square degrees.
        zBinEdges (:obj:`np.ndarray`): Defines the redshift bins for the cluster counts.
        z (:obj:`np.ndarray`): Centers of the redshift bins.
        log10M (:obj:`np.ndarray`): Centers of the log10 mass bins for the cluster counts
            (in MSun, with mass defined according to `delta` and `rhoType`).
        a (:obj:`np.ndarray`): Scale factor (1/(1+z)).
        delta (:obj:``float`): Overdensity parameter, used for mass definition (e.g., 200, 500).
        rhoType (:obj:`str`): Density definition, either 'matter' or 'critical', used for mass definition.
        mdef (:obj:`pyccl.halos.massdef.MassDef`): CCL mass definition object, defined by `delta` and `rhoType`.
        transferFunction (:obj:`str`): Transfer function to use, as understood by CCL (e.g., 'eisenstein_hu', 
            'boltzmann_camb').
        H0 (:obj:`float`): The Hubble constant at redshift 0, in km/s/Mpc.
        Om0 (:obj:`float`): Dimensionless total (dark + baryonic) matter density parameter at redshift 0.
        Ob0 (:obj:`float`): Dimensionless baryon matter density parameter at redshift 0.
        sigma8 (:obj:`float`): Defines the amplitude of the matter power spectrum.
        ns (:obj:`float`): Scalar spectral index of matter power spectrum.
        volumeMpc3 (:obj:`float`): Co-moving volume in Mpc3 for the given survey area and cosmological
            parameters.
        numberDensity (:obj:`np.ndarray`): Number density of clusters (per cubic Mpc) on the 
            (z, log10M) grid.
        clusterCount (:obj:`np.ndarray`): Cluster counts on the (z, log10M) grid.
        numClusters (:obj:`float`): Total number of clusters in the survey area above the minimum mass
            limit.
        numClustersByRedshift (:obj:`np.ndarray`): Number of clusters in the survey area above the
            minimum mass limit, as a function of redshift.
    
    """
    def __init__(self, minMass, maxMass, areaDeg2, zMin, zMax, H0, Om0, Ob0, sigma8, ns, zStep = 0.01,
                 numMassBins = 20000, enableDrawSample = False, delta = 500, rhoType = 'critical',
                 transferFunction = 'boltzmann_camb', massFunction = 'Tinker08',
                 c_m_relation = 'Bhattacharya13'):
        """Create a MockSurvey object, for performing calculations of cluster counts or generating mock
        catalogs. The Tinker et al. (2008) halo mass function is used (hardcoded at present, but in 
        principle this can easily be swapped for any halo mass function supported by CCL).
        
        Args:
            minMass (:obj:`float`): The minimum mass, in MSun. This should be set considerably lower than
                the actual survey completeness limit, otherwise completeness calculations will be wrong.
            areaDeg2 (:obj:`float`): Specifies the survey area in square degrees, which scales the
                resulting cluster counts accordingly.
            zMin (:obj:`float`): Minimum redshift for the (z, log10M) grid.
            zMax (:obj:`float`): Maximum redshift for the (z, log10M) grid.
            H0 (:obj:`float`): The Hubble constant at redshift 0, in km/s/Mpc.
            Om0 (:obj:`float`): Dimensionless total (dark + baryonic) matter density parameter at redshift 0.
            Ob0 (:obj:`float`): Dimensionless baryon matter density parameter at redshift 0.
            sigma8 (:obj:`float`): Defines the amplitude of the matter power spectrum.
            ns (:obj:`float`): Scalar spectral index of matter power spectrum.  
            zStep (:obj:`float`, optional): Sets the linear spacing between redshift bins.
            enableDrawSample (:obj:`bool`, optional): This needs to be set to True to enable use of the
                :func:`self.drawSample` function. Setting this to False avoids some overhead.
            delta (:obj:``float`): Overdensity parameter, used for mass definition (e.g., 200, 500).
            rhoType (:obj:`str`): Density definition, either 'matter' or 'critical', used for mass definition.
            transferFunction (:obj:`str`): Transfer function to use, as understood by CCL (e.g., 'eisenstein_hu', 
                'boltzmann_camb').
            massFunction (:obj:`str`): Name of the mass function to use, currently either 'Tinker08' or
                'Tinker10'. Mass function calculations are done by CCL.
            c_m_relation ('obj':`str`): Name of the concentration -- mass relation to assume, as understood by
                CCL (this may be used internally for conversion between mass definitions, as needed).

        """
        
        if areaDeg2 == 0:
            raise Exception("Cannot create a MockSurvey object with zero area")
        self.areaDeg2=areaDeg2
        self.areaSr=np.radians(np.sqrt(areaDeg2))**2
        self.fsky=self.areaSr/(4*np.pi)

        self.delta=delta
        self.rhoType=rhoType
        self.c_m_relation=c_m_relation
        self.mdef=ccl.halos.MassDef(self.delta, self.rhoType)
        self.transferFunction=transferFunction
        self.massFuncName=massFunction
        
        self.log10M=np.arange(np.log10(minMass), np.log10(maxMass), 0.001)
        
        class_sz_cosmo_params = {
        'Omega_b': Ob0,
        'Omega_cdm':  Om0-Ob0,
        'H0': H0,
        'sigma8': sigma8,
        'tau_reio':  0.0561, ## doesnt matter 
        'n_s': ns,
        }
        
        # tenToA0, B0, Mpivot, sigma_int = [scalingRelationDict['tenToA0'],
        #                                   scalingRelationDict['B0'],
        #                                   scalingRelationDict['Mpivot'],
        #                                   scalingRelationDict['sigma_int']]
        #
        # class_sz_ym_params = {
        # 'A_ym'  : tenToA0,
        # 'B_ym'  : B0,
        # 'C_ym' : 0.,
        # 'sigmaM_ym' : sigma_int,
        # 'm_pivot_ym_[Msun]' : Mpivot,
        # }
        # #BB


        self.cosmo = Class()
        self.cosmo.set(common_class_sz_settings)
        self.cosmo.set(class_sz_cosmo_params)
        # self.cosmo.set(class_sz_ym_params)
        self.cosmo.compute_class_szfast()
        

        zRange=np.arange(zMin, zMax+zStep, zStep)
        self.zBinEdges=zRange
        self.z=(zRange[:-1]+zRange[1:])/2.
        self.a=1./(1+self.z)
        nzs=self.z.shape[0]

        # NOTE: We use MSun, Boris/class-sz uses MSun/h, watch out for that
        self.log10M=np.linspace(np.log10(minMass), np.log10(maxMass), numMassBins)
        self.M=np.power(10, self.log10M)
        self.log10MBinEdges=np.linspace(self.log10M.min()-(self.log10M[1]-self.log10M[0])/2,
                                        self.log10M.max()+(self.log10M[1]-self.log10M[0])/2, len(self.log10M)+1)

        # Below is needed for Q calc when not using M500c definition (for now at least)
        if (self.delta == 500 and self.rhoType == 'critical') == False:
            self._M500cDef=ccl.halos.MassDef(500, "critical")
            self._transToM500c=ccl.halos.mass_translator(mass_in = self.mdef, mass_out = self._M500cDef,
                                                         concentration = self.c_m_relation)

        self.enableDrawSample=enableDrawSample
        self.update(H0, Om0, Ob0, sigma8, ns)
        print("done __init__")


    def update(self, H0, Om0, Ob0, sigma8, ns):
        """Recalculate cluster counts for the updated cosmological parameters given.

        Args:
            H0 (:obj:`float`): The Hubble constant at redshift 0, in km/s/Mpc.
            Om0 (:obj:`float`): Dimensionless total (dark + baryonic) matter density parameter at redshift 0.
            Ob0 (:obj:`float`): Dimensionless baryon matter density parameter at redshift 0.
            sigma8 (:obj:`float`): Defines the amplitude of the matter power spectrum.
            ns (:obj:`float`): Scalar spectral index of matter power spectrum.

        """

        print("in update")
        self._get_new_cosmo(H0, Om0, Ob0, sigma8, ns)

        self._doClusterCount()

        # CCL
        # For quick Q, fRel calc (these are in MockSurvey rather than SelFn as used by drawSample)
#         self.theta500Splines=[]
#         self.fRelSplines=[]
#         self.Ez=ccl.h_over_h0(self.cosmoModel,self.a)
#         self.Ez2=np.power(self.Ez, 2)
#         self.DAz=ccl.angular_diameter_distance(self.cosmoModel,self.a)
#         self.criticalDensity=ccl.physical_constants.RHO_CRITICAL*(self.Ez*self.cosmoModel['h'])**2
#         for k in range(len(self.z)):
#             # NOTE: Q fit uses theta500, as does fRel (hardcoded M500 - T relation in there)
#             # This bit here may not be strictly necessary, since we don't need to map on to binning
#             if self.delta == 500 and self.rhoType == "critical":
#                 interpLim_minLog10M500c=self.log10M.min()
#                 interpLim_maxLog10M500c=self.log10M.max()
#             else:
#                 interpLim_minLog10M500c=np.log10(self._transToM500c(self.cosmoModel, self.M.min(), self.a[k]))
#                 interpLim_maxLog10M500c=np.log10(self._transToM500c(self.cosmoModel, self.M.max(), self.a[k]))
#             zk=self.z[k]
#             interpPoints=100
#             fitM500s=np.power(10, np.linspace(interpLim_minLog10M500c, interpLim_maxLog10M500c, interpPoints))
#             fitTheta500s=np.zeros(len(fitM500s))
#             fitFRels=np.zeros(len(fitM500s))
#             criticalDensity=self.criticalDensity[k]
#             DA=self.DAz[k]
#             Ez=self.Ez[k]
#             R500Mpc=np.power((3*fitM500s)/(4*np.pi*500*criticalDensity), 1.0/3.0)
#             fitTheta500s=np.degrees(np.arctan(R500Mpc/DA))*60.0
#             fitFRels=signals.calcFRel(zk, fitM500s, Ez)
#             tckLog10MToTheta500=interpolate.splrep(np.log10(fitM500s), fitTheta500s)
#             tckLog10MToFRel=interpolate.splrep(np.log10(fitM500s), fitFRels)
#             self.theta500Splines.append(tckLog10MToTheta500)
#             self.fRelSplines.append(tckLog10MToFRel)
#
#         # Stuff to enable us to draw mock samples (see drawSample)
#         # Interpolators here need to be updated each time we change cosmology
#         if self.enableDrawSample == True:
#
#             # For drawing from overall z distribution
#             zSum=self.clusterCount.sum(axis = 1)
#             pz=np.cumsum(zSum)/self.numClusters
#             self.zRoller=_spline(pz, self.z, k = 3)
#
#             # For drawing from each log10M distribution at each point on z grid
#             # And quick fRel, Q calc using interpolation
#             # And we may as well have E(z), DA on the z grid also
#             self.log10MRollers=[]
#             for i in range(len(self.z)):
#                 ngtm=self._cumulativeNumberDensity(self.z[i])
#                 mask=ngtm > 0
#                 self.log10MRollers.append(_spline((ngtm[mask] / ngtm[0])[::-1], np.log10(self.M[mask][::-1]), k=3))


    def _get_new_cosmo(self, H0, Om0, Ob0, sigma8, ns):
        pass
        # # CCL
        # if ((self.H0 != H0) or (self.Om0 != Om0) or
        #     (self.Ob0 != Ob0) or (self.sigma8 != sigma8)):
        #     self.H0=H0
        #     self.Om0=Om0
        #     self.Ob0=Ob0
        #     self.sigma8=sigma8
        #     self.ns=ns
        #     self.cosmoModel=ccl.Cosmology(Omega_c=Om0-Ob0,
        #                                   Omega_b=Ob0,
        #                                   h=0.01*H0,
        #                                   sigma8=sigma8,
        #                                   n_s=ns,
        #                                   transfer_function=self.transferFunction)
        #     if self.massFuncName == 'Tinker10':
        #         self.mfunc=ccl.halos.MassFuncTinker10(mass_def = self.mdef)
        #     elif self.massFuncName == 'Tinker08':
        #         self.mfunc=ccl.halos.MassFuncTinker08(mass_def = self.mdef)


    def _doClusterCount(self):
        """Updates cluster count etc. after mass function object is updated.

        """

        print("in _doClusterCount")
        assert(self.areaSr == np.radians(np.sqrt(self.areaDeg2))**2)
        self.fsky=self.areaSr/(4*np.pi) # in case it was updated

        # CLASS-SZ
        # Still need to implement (maybe): self.volumeMpc3, self.numberDensity
        lnms=np.log(np.power(10,  self.log10M)*self.cosmo.h())
        dndmdz = np.zeros((self.log10M.shape[0], self.z.shape[0]))
        for (im,mm) in enumerate(lnms):
            dndmdz[im,:] = 4.*np.pi*self.fsky*np.vectorize(self.cosmo.get_volume_dVdzdOmega_at_z)(self.z)*np.vectorize(self.cosmo.get_dndlnM_at_z_and_M)(self.z,np.exp(mm))
        self.HMFRange=np.array([np.min(dndmdz),np.max(dndmdz)])
        dndz=np.trapz(dndmdz,x = lnms,axis = 0)
        self.numClusters=np.trapz(dndz, x = self.z)
        self.numClustersByRedshift=self.numClustersByRedshift=dndz*np.gradient(self.z)
        # I think below should be good to 0.1%
        norm=np.sum(dndmdz)/self.numClusters
        self.clusterCount=dndmdz/norm

        # # CCL
        # # Number density by z and total cluster count (in redshift shells)
        # zRange=self.zBinEdges
        # h = self.cosmoModel['h']
        # self.M=np.power(10, self.log10M) # in M_sun
        # norm_mfunc=1. / np.log(10)
        # numberDensity=[]
        # clusterCount=[]
        # totalVolumeMpc3=0.
        # for i in range(len(zRange)-1):
        #     zShellMin=zRange[i]
        #     zShellMax=zRange[i+1]
        #     zShellMid=(zShellMax+zShellMin)/2.
        #     dndlnM=self.mfunc(self.cosmoModel, self.M, 1./(1+zShellMid)) * norm_mfunc
        #     dndM = dndlnM / self.M
        #     n=dndM * np.gradient(self.M)
        #     numberDensity.append(n)
        #     shellVolumeMpc3=self._comovingVolume(zShellMax)-self._comovingVolume(zShellMin)
        #     shellVolumeMpc3=shellVolumeMpc3*(self.areaSr/(4*np.pi))
        #     totalVolumeMpc3+=shellVolumeMpc3
        #     clusterCount.append(n*shellVolumeMpc3)
        # numberDensity=np.array(numberDensity)
        # clusterCount=np.array(clusterCount)
        # self.volumeMpc3=totalVolumeMpc3
        # self.numberDensity=numberDensity
        # self.clusterCount=clusterCount
        # self.numClusters=np.sum(clusterCount)
        # self.numClustersByRedshift=np.sum(clusterCount, axis = 1)

    def drawSample(self, y0Noise, scalingRelationDict, QFit = None, wcs = None, photFilterLabel = None,\
                   tileName = None, SNRLimit = None, makeNames = False, z = None, numDraws = None,\
                   areaDeg2 = None, applySNRCut = False, applyPoissonScatter = True,\
                   applyIntrinsicScatter = True, applyNoiseScatter = True,\
                   applyRelativisticCorrection = True, verbose = False, biasModel = None):
        """Draw a cluster sample from the mass function, generating mock y0~ values (called `fixed_y_c` in
        Nemo catalogs) by applying the given scaling relation parameters, and then (optionally) applying
        a survey selection function.

        Args:
            y0Noise (:obj:`float` or :obj:`np.ndarray`): Either a single number (if using e.g., a survey
                average), an RMS table (with columns 'areaDeg2' and 'y0RMS'), or a noise map (2d array).
                A noise map must be provided here if you want the output catalog to contain RA, dec
                coordinates (in addition, a WCS object must also be provided - see below).
            scalingRelationDict (:obj:`dict`): A dictionary containing keys 'tenToA0', 'B0', 'Mpivot',
                'sigma_int' that describes the scaling relation between y0~ and mass (this is the
                format of `massOptions` in Nemo .yml config files).
            QFit (:obj:`nemo.signals.QFit`, optional): Object that handles the filter mismatch
                function, *Q*. If not given, the output catalog will not contain `fixed_y_c` columns,
                only `true_y_c` columns.
            wcs (:obj:`astWCS.WCS`, optional): WCS object corresponding to `y0Noise`, if `y0Noise` is
                as noise map (2d image array). Needed if you want the output catalog to contain RA, dec
                coordinates.
            photFilterLabel (:obj:`str`, optional): Name of the reference filter (as defined in the
                Nemo .yml config file) that is used to define y0~ (`fixed_y_c`) and the filter mismatch
                function, Q.
            tileName (:obj:`str`, optional): Name of the tile for which the sample will be generated.
            SNRLimit (:obj:`float`, optional): Signal-to-noise detection threshold used for the
                output catalog (corresponding to a cut on `fixed_SNR` in Nemo catalogs). Only applied
                if `applySNRCut` is also True (yes, this can be cleaned up).
            makeNames (:obj:`bool`, optional): If True, add names of the form MOCK CL JHHMM.m+/-DDMM
                to the output catalog.
            z (:obj:`float`, optional): If given produce a sample at the nearest z in the MockSurvey
                z grid. The default behaviour is to use the full redshift grid specified by `self.z`.
            numDraws (:obj:`int`, optional): If given, the number of draws to perform from the mass
                function, divided equally among the redshift bins. The default is to use the values
                contained in `self.numClustersByRedshift`.
            areaDeg2 (:obj:`float`, optional): If given, the cluster counts will be scaled to this
                area. Otherwise, they correspond to `self.areaDeg2`. This parameter will be ignored
                if `numDraws` is also given.
            applySNRCut (:obj:`bool`, optional): If True, cut the output catalog according to the
                `fixed_SNR` threshold set by `SNRLimit`.
            applyPoissonScatter (:obj:`bool`, optional): If True, add Poisson noise to the cluster
                counts (implemented by modifiying the number of draws from the mass function).
            applyIntrinsicScatter (:obj:`bool`, optional): If True, apply intrinsic scatter to the
                SZ measurements (`fixed_y_c`), as set by the `sigma_int` parameter in
                `scalingRelationDict`.
            applyNoiseScatter (:obj:`bool`, optional): If True, apply measurement noise, generated
                from the given noise level or noise map (`y0Noise`), to the output SZ measurements
                (`fixed_y_c`).
            applyRelativisticCorrection (:obj:`bool`, optional): If True, apply the relativistic
                correction.

        Returns:
            A catalog as an :obj:`astropy.table.Table` object, in the same format as produced by
            the main `nemo` script.

        Notes:
            If both `applyIntrinsicScatter`, `applyNoiseScatter` are set to False, then the output
            catalog `fixed_y_c` values will be exactly the same as `true_y_c`, although each object
            will still have an error bar listed in the output catalog, corresponding to its location
            in the noise map (if given).

        """

        # self.n_tot_obs = np.random.poisson(lam=th_ntot)

        # print('>>> ntot obs for this catalog: ', self.n_tot_obs)

        print("in drawSample()")
        import IPython
        IPython.embed()
        sys.exit()

        if numDraws is not None:
            numClusters=numDraws

        if applyPoissonScatter == True:
            numClusters=np.random.poisson(lam = self.numClusters)
        else:
            numClusters=int(round(self.numClusters))

        # If given y0Noise as RMSMap, draw coords (assuming clusters aren't clustered - which they are...)
        # NOTE: switched to using valid part of RMSMap here rather than areaMask - we need to fix the latter to same area
        # It isn't a significant issue though
        # This generates even density RA, dec coords on the whole sky taking into account the projection
        # Consequently, this is inefficient if fed individual tiles rather than a full sky noise map
        RMSMap=y0Noise
        xsList=[]
        ysList=[]
        maxCount=100000
        count=0
        while(len(xsList) < numClusters):
            count=count+1
            if count > maxCount:
                raise Exception("Failed to generate enough random coords in %d iterations" % (maxCount))
            theta=np.degrees(np.pi*2*np.random.uniform(0, 1, numClusters))
            phi=np.degrees(np.arccos(2*np.random.uniform(0, 1, numClusters)-1))-90
            xyCoords=np.array(wcs.wcs2pix(theta, phi))
            xs=np.array(np.round(xyCoords[:, 0]), dtype = int)
            ys=np.array(np.round(xyCoords[:, 1]), dtype = int)
            mask=np.logical_and(np.logical_and(xs >= 0, xs < RMSMap.shape[1]), np.logical_and(ys >= 0, ys < RMSMap.shape[0]))
            xs=xs[mask]
            ys=ys[mask]
            mask=RMSMap[ys, xs] > 0
            xsList=xsList+xs[mask].tolist()
            ysList=ysList+ys[mask].tolist()
        xs=np.array(xsList)[:numClusters]
        ys=np.array(ysList)[:numClusters]
        del xsList, ysList
        RADecCoords=wcs.pix2wcs(xs, ys)
        RADecCoords=np.array(RADecCoords)
        RAs=RADecCoords[:, 0]
        decs=RADecCoords[:, 1]
        y0Noise=RMSMap[ys, xs]

        # CLASS-SZ: draw zs and masses using rejection sampling (from Inigo Zubeldia)
        zs=np.zeros(y0Noise.shape)
        zErrs=np.zeros(y0Noise.shape)
        log10Ms=np.zeros(y0Noise.shape)
        for clusterIndex in range(numClusters):
            hmfEval=0.
            hmfSample=1.
            while hmfSample > hmfEval:
                ln10MSample=np.random.rand()*(self.log10M[-1]-self.log10M[0])+self.log10M[0]
                lnMhSample=np.log(10**ln10m_sample*self.cosmo.h())
                zSample=np.random.rand()*(self.z[-1]-self.z[0])+self.z[0]
                hmfSample=np.random.rand()*(self.hmf_range[-1]-self.hmf_range[0])+self.hmf_range[0]
                hmfEval=4.*np.pi*self.fsky*self.cosmo.get_volume_dVdzdOmega_at_z(z_sample)*self.cosmo.get_dndlnM_at_z_and_M(z_sample,np.exp(lnmh_sample))
            log10Ms[cluster_index] = ln10m_sample
            zs[cluster_index] = z_sample
        
        #
                   

        names=[]
        for RADeg, decDeg in zip(RAs, decs):
            names.append(catalogs.makeName(RADeg, decDeg, prefix = 'MOCK-CL'))
                
        
        
        M200c = np.power(10, log10Ms)
        true_y0s = np.vectorize(self.cosmo.get_y_at_m_and_z)(M200c*self.cosmo.h(),zs)
        
        tab = atpy.Table()
        
        tab.add_column(atpy.Column(names, 'name'))
        tab.add_column(atpy.Column(RAs, 'RADeg'))
        tab.add_column(atpy.Column(decs, 'decDeg'))
        tab.add_column(atpy.Column(np.power(10, log10Ms)/1e14, "true_M200c"))
        tab.add_column(atpy.Column(true_y0s/1e-4, 'true_y_c'))
        tab.add_column(atpy.Column(zs, 'redshift'))
        tab.add_column(atpy.Column(zErrs, 'redshiftErr'))
                

        return tab



"""

Fit the scaling relation in the sims

"""

import os
import sys
import numpy as np
import astropy.table as atpy
from nemo import catalogs, signals, plotSettings, MockSurvey
from astropy.cosmology import FlatLambdaCDM
from scipy import stats
import pyccl as ccl
import pylab as plt
import IPython

#------------------------------------------------------------------------------------------------------------
def calcMass(tab, massOptions, tckQFitDict, fRelWeightsDict, mockSurvey):
    """Calculates masses for cluster data in table. Because why not re-do on the fly when hippo busy?
    
    """
    
    count=0
    for row in tab:
        count=count+1
        #print("... %d/%d; %s (%.3f +/- %.3f) ..." % (count, len(tab), row['name'], 
                                                              #row['redshift'], row['redshiftErr']))

        tileName=row['tileName']
        
        # Cuts on z, fixed_y_c for forced photometry mode (invalid objects will be listed but without a mass)
        if row['fixed_y_c'] > 0 and np.isnan(row['redshift']) == False:
            # Corrected for mass function steepness
            massDict=signals.calcMass(row['fixed_y_c']*1e-4, row['fixed_err_y_c']*1e-4, 
                                      row['redshift'], row['redshiftErr'],
                                      tenToA0 = massOptions['tenToA0'],
                                      B0 = massOptions['B0'], 
                                      Mpivot = massOptions['Mpivot'], 
                                      sigma_int = massOptions['sigma_int'],
                                      tckQFit = tckQFitDict[tileName], mockSurvey = mockSurvey, 
                                      applyMFDebiasCorrection = True,
                                      applyRelativisticCorrection = True,
                                      fRelWeightsDict = fRelWeightsDict[tileName])
            row['M500c']=massDict['M500c']
            try:
                row['M500c_errPlus']=massDict['M500c_errPlus']
            except:
                IPython.embed()
                sys.exit()
            row['M500c_errMinus']=massDict['M500c_errMinus']

    return tab

#------------------------------------------------------------------------------------------------------------
# Main

# Options - masses are as output by Nemo routines; we compare these to halo catalog (converting as necessary)
massCol="M500c"
#massCol="M200m"

# Websky cosmo - for on-the-fly redone masses
minMass=1e13
areaDeg2=700.   # doesn't matter
zMin=0.0
zMax=2.0
H0, Om0, Ob0, sigma8, ns = 68.0, 0.049+0.261, 0.049, 0.81, 0.965
TCMB=2.72548
cosmoModel=FlatLambdaCDM(H0 = H0, Om0 = Om0, Ob0 = Ob0, Tcmb0 = TCMB)
mockSurvey=MockSurvey.MockSurvey(minMass, areaDeg2, zMin, zMax, H0, Om0, Ob0, sigma8, ns)
massOptions={'tenToA0': 2.65e-05,
             'B0': 0.0,
             'Mpivot': 3.0e+14, 
             'sigma_int': 0.2}
tckQFitDict=signals.loadQ("../MFMF_SOSim_3freq_tiles/selFn/QFit.fits")
fRelWeightsDict=signals.loadFRelWeights("../MFMF_SOSim_3freq_tiles/selFn/fRelWeights.fits")

# Make combined table
mergedTabFileName="trueMasses_MFMF_SOSim_3freq_tiles_M500c.fits"
if os.path.exists(mergedTabFileName) == False:
    halos=atpy.Table().read("../halos.fits")
    tab=atpy.Table().read("../MFMF_SOSim_3freq_tiles/MFMF_SOSim_3freq_tiles_mass.fits")
    tab=tab[tab['fixed_SNR'] > 6]

    tab, halos, rDeg=catalogs.crossMatch(tab, halos, radiusArcmin = 1.0)

    zs=halos['z']
    yc=tab['fixed_y_c']
    M200m=halos['M200m']
    
    M200mDef=ccl.halos.MassDef200m(c_m='Bhattacharya13')
    M500cDef=ccl.halos.MassDef(500, "critical")
    M500c=[]
    count=0
    for m, z in zip(M200m, zs):
        M500c.append(M200mDef.translate_mass(mockSurvey.cosmoModel, m, 1/(1+z), M500cDef))
    M500c=np.array(M500c)
    
    tab['true_M500c']=M500c/1e14
    tab['true_M200m']=M200m/1e14
    tab['redshift']=zs
    tab.write(mergedTabFileName, overwrite = True)

# Re-do masses on the fly
tab=atpy.Table().read(mergedTabFileName)

# Cut on mass and z to do the fit
MMin=3.0
zBinEdges=[0.2, 0.4, 0.6, 0.8, 1.0]
for i in range(len(zBinEdges)-1):
    zMin=zBinEdges[i]
    zMax=zBinEdges[i+1]

    fitTab=tab[tab['M500c'] > MMin]
    fitTab=fitTab[fitTab['redshift'] > zMin]
    fitTab=fitTab[fitTab['redshift'] < zMax]

    # NOTE: This is done in place anyway
    fitTab=calcMass(fitTab, massOptions, tckQFitDict, fRelWeightsDict, mockSurvey)

    y=fitTab['M500c']
    x=fitTab['true_M500c']
    result=stats.linregress(x, y)
    sumSqRes=np.sum((x-y)**2)
    calibFactor=np.mean(fitTab['true_M500c'])/np.mean(fitTab['M500c'])
    
    # Scaling relation plot
    plotSettings.update_rcParams()
    plt.figure(figsize=(9.5,9))
    ax=plt.axes([0.1, 0.1, 0.89, 0.89])
    ax.set_aspect('equal')        
    plotRange=np.linspace(1.0, 50.0, 100)
    plt.plot(x, y, '.')
    plt.plot(plotRange, plotRange, 'k-')
    plt.xlabel("$M^{\\rm true}_{\\rm 500c}$ (10$^{14}$ $M_{\odot}$)")
    plt.ylabel("$M_{\\rm 500c}$ (10$^{14}$ $M_{\odot}$)")
    plt.xlim(2, 50)
    plt.ylim(2, 50)
    plt.loglog()
    plt.title("%.1f < z < %.1f" % (zMin, zMax))
    plt.savefig("massRecovery_%.1f_%.1f.png" % (zMin, zMax))
    plt.close()

    print("%.1f < z < %.1f:" % (zMin, zMax))
    print("    calibFactor = ", calibFactor)
    print("    sumSqRes = ", sumSqRes)

    #IPython.embed()
    #sys.exit()



"""

Reads in the halos.pksc file and outputs .fits and .reg format catalogs for 
the 1 million most massive halos (M200m > 1e14 MSun). Cuts down to SO dec 
range.

"""
import os
import sys
import healpy as hp
import astropy.table as atpy
from nemo import catalogs
from cosmology import *
import numpy as np
import IPython

rho = 2.775e11*omegam*h**2 # Msun/Mpc^3

f=open('../WebSky/halos.pksc', 'rb')
N=np.fromfile(f,count=3,dtype=np.int32)[0]

# only take first five entries for testing (there are ~8e8 halos total...)
# comment the following line to read in all halos
N = 1000000

catalog=np.fromfile(f,count=N*10,dtype=np.float32)
catalog=np.reshape(catalog,(N,10))

x  = catalog[:,0];  y = catalog[:,1];  z = catalog[:,2] # Mpc (comoving)
vx = catalog[:,3]; vy = catalog[:,4]; vz = catalog[:,5] # km/sec
R  = catalog[:,6] # Mpc

# convert to mass, comoving distance, radial velocity, redshfit, RA and DEc
M        = 4*np.pi/3.*rho*R**3        # Msun (M200m)
chi      = np.sqrt(x**2+y**2+z**2)    # Mpc
vrad     = (x*vx + y*vy + z*vz) / chi # km/sec
redshift = zofchi(chi)      

theta, phi  = hp.vec2ang(np.column_stack((x,y,z))) # in radians

decDeg=-1*(np.degrees(theta)-90) # Because HEALPix
RADeg=np.degrees(phi)

names=[]
for ra, dec in zip(RADeg, decDeg):
    names.append(catalogs.makeACTName(ra, dec, prefix= 'MOCK-CL'))

outFileName="halos.fits"
tab=atpy.Table()
tab.add_column(atpy.Column(names, 'name'))
tab.add_column(atpy.Column(RADeg, 'RADeg'))
tab.add_column(atpy.Column(decDeg, 'decDeg'))
tab.add_column(atpy.Column(M, 'M200m'))
tab.add_column(atpy.Column(redshift, "z"))
tab=tab[np.where(tab['decDeg'] < 30)]
tab=tab[np.where(tab['decDeg'] > -70)]
tab.write(outFileName, overwrite = True)

catalogs.catalog2DS9(tab, "halos.reg")

#IPython.embed()
#sys.exit()

### e.g. project to a map, matching the websky orientations
#nside = 1024
#map   = np.zeros((hp.nside2npix(nside)))

#pix = hp.vec2pix(nside, x, y, z)
#pix = hp.ang2pix(nside, theta, phi) does the same

#weight = 1. #1 for number density, array of size(x) for arbitrary
#np.add.at(map, pix, weight)



"""

N(z) check - for comparing across codes using different mass function implementations etc..

NOTE: We messed up when setting this task, so the mass limit is M500c > 5e13 MSun/h

"""

import os
import sys
import numpy as np
import pylab as plt
import astropy.table as atpy
from astLib import *
from scipy import stats
from scipy import interpolate
from nemo import completeness
import argparse
import time
import IPython
   
#------------------------------------------------------------------------------------------------------------
if __name__ == '__main__':

    parser=argparse.ArgumentParser("NzCheck.py")
    parser.add_argument("selFnDir", help="""Directory containing files needed for computing the selection 
                        function.""")
    parser.add_argument("-c", "--config", dest = "configFileName", help="""A .yml configuration file. If
                        this is not given, the config.yml file in selFnDir will be used.""",
                        default = None)
    parser.add_argument("-f", "--footprint", dest = "footprint", help="""Footprint to use, e.g., DES,
                        HSC, KiDS (default: full).""", default = None)
    parser.add_argument("-S", "--SNR-cut", dest = "SNRCut", help="""Use only clusters with fixed_SNR > 
                        this value.""", default = 5.0, type = float)
    parser.add_argument("-z", "--z-step", dest = "zStep", help="""Redshift bin width (default: 0.1).""", 
                        default = 0.1, type = float)
    args = parser.parse_args()
    
    configFileName=args.configFileName
    selFnDir=args.selFnDir
    SNRCut=args.SNRCut
    zStep=args.zStep
    footprintLabel=args.footprint
    
    print(">>> Setting up SNR > %.2f selection function ..." % (SNRCut))
    selFn=completeness.SelFn(selFnDir, SNRCut, configFileName = configFileName, 
                             footprintLabel = footprintLabel, zStep = zStep)

    scalingRelationDict=selFn.scalingRelationDict
    H0, Om0, Ob0, sigma8, ns = 70.0, 0.30, 0.05, 0.80, 0.95
    h=H0/100.0
    selFn.update(H0, Om0, Ob0, sigma8, ns, scalingRelationDict = scalingRelationDict)
    print("Total area = %.3f square degrees" % (selFn.totalAreaDeg2))
    
    # N(z) with M500c > 5e13 MSun - no selection function applied
    countsByRedshift=selFn.mockSurvey.clusterCount[:, np.greater(selFn.mockSurvey.log10M, np.log10(5e13/h))].sum(axis = 1)
    with open("NzCheck_noSelFn.csv", "w") as outFile:
        for i in range(len(selFn.mockSurvey.z)):
            outFile.write("%.1f <= z < %.1f\t%.3f\n" % (selFn.mockSurvey.zBinEdges[i], selFn.mockSurvey.zBinEdges[i+1], countsByRedshift[i]))
    
    # N(z) with M500c > 5e13 MSun - with S/N > 5 selection function applied
    predMz=selFn.compMz*selFn.mockSurvey.clusterCount
    countsByRedshift=predMz[:, np.greater(selFn.mockSurvey.log10M, np.log10(5e13/h))].sum(axis = 1)
    with open("NzCheck_withSelFn.csv", "w") as outFile:
        for i in range(len(selFn.mockSurvey.z)):
            outFile.write("%.1f <= z < %.1f\t%.3f\n" % (selFn.mockSurvey.zBinEdges[i], selFn.mockSurvey.zBinEdges[i+1], countsByRedshift[i]))

       



"""

nemo - SZ cluster and source detection package

"""

from . import catalogs 
from . import photometry
from . import plotSettings
from . import startUp
from . import filters
from . import signals
from . import completeness
from . import maps
from . import pipelines

__all__ = ['startUp', 'catalogs', 'maps', 'filters', 'photometry', 
           'signals', 'completeness', 'pipelines', 'plotSettings']

from ._version import get_versions
__version__ = get_versions()['version']
del get_versions
